/* $Id: buffer.c,v 1.6 2008/06/21 06:53:27 maxim Exp $
 *
 */

#include <stdlib.h>

#include "configure.h"

#include "common/misc.h"
#include "httpd.h"
#include "buffer.h"

#define MAX_SIZES (MAXIMUM_BUFFER_SIZE / INITIAL_BUFFER_SIZE)
#define BUFFERS_STEP 128*1024

static Buffer* freeBuffers[MAX_SIZES];
static int nBufferBlocks = 0;

Buffer*
newBuffer(int size)
{
	Buffer* buffer;
	int s;
	debug("newBuffer(%d)", size);

	if (size > MAXIMUM_BUFFER_SIZE ||
		(size % INITIAL_BUFFER_SIZE) ||
		countBits(s = size / INITIAL_BUFFER_SIZE) != 1) {
		if ((buffer = MALLOC(sizeof(Buffer) + size))) {
			buffer->buffer = (char*)(buffer + 1);
			buffer->size = size;
			buffer->start = buffer->count = 0;
		}
		return buffer;
	}

	if (!(buffer = freeBuffers[--s])) {
		int i;
		int count = BUFFERS_STEP / size;
		debug("newBuffer: allocating buffers-%d N%d (%dx%d bytes)", s, nBufferBlocks, count, size);
		freeBuffers[s] = MALLOC(count * (sizeof(Buffer) + size));
		if (!freeBuffers[s]) return NULL;
		for (i = 0, buffer = freeBuffers[s]; i < count - 1; i++) {
			buffer->buffer = (void*)buffer + sizeof(Buffer) + size;
			buffer = (Buffer*)buffer->buffer;
		}
		buffer->buffer = NULL;
		buffer = freeBuffers[s];
		nBufferBlocks++;
		//buffer = bufferBlocks[nBufferBlocks++] = freeBuffers[s];
	}

	freeBuffers[s] = (Buffer*)buffer->buffer;

	buffer->buffer = (char*)(buffer + 1);
	buffer->size = size;
	buffer->start = buffer->count = 0;
	debug("buffer: %p (%d), data: %p", buffer, sizeof(Buffer), buffer->buffer);
	return buffer;
}

Buffer*
freeBuffer(Buffer* buffer)
{
	int s;
	debug("freeBuffer(%p): %d bytes", buffer, buffer ? buffer->size : 0);
	if (!buffer) return NULL;

	if (buffer->size > MAXIMUM_BUFFER_SIZE ||
		(buffer->size % INITIAL_BUFFER_SIZE) ||
		countBits(s = buffer->size / INITIAL_BUFFER_SIZE) != 1)
		return FREE(buffer);

	buffer->buffer = (char*)freeBuffers[--s];
	freeBuffers[s] = buffer;
	return NULL;
}


#if 0
void
freeAllBuffers()
{
	int s;
	for (s = 0; s < nBufferBlocks; s++)
		FREE(bufferBlocks[s]);

	for (s = 0; s < MAX_SIZES; s++) {
		freeBuffers[s] = NULL;
		buffersSizes[s] = 0;
	}

}
#endif

