/* $Id: core.c,v 1.54 2008/09/20 23:04:02 maxim Exp $
 *
 */

#define _GNU_SOURCE

#include "configure.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/uio.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>


#include "common/settings.h"
#include "common/signames.h"
#include "common/ip_ranges.h"
#include "common/misc.h"
#include "common/str.h"
#include "core.h"
#include "httpd.h"
#include "flood.h"
#include "proxy.h"
#include "config.h"
#include "handlers.h"
#include "files.h"
#include "buffer.h"
#include "access.h"

#if DEBUGLOG
#include "common/ptrinfo.h"
#endif


#if SUPPORT_CACHE
#include "cache.h"
#endif


#if TRANSFER_MODE == TRANSFER_SENDFILE

#if defined(HAS_LINUX_SENDFILE)
#include <sys/sendfile.h>
#endif

#endif /* TRANSFER_MODE == TRANSFER_SENDFILE */


#if MODEL == MODEL_KEVENT
#include <sys/event.h>
#endif
#if MODEL == MODEL_EPOLL
#include <sys/epoll.h>
#endif

static char *states[] = {
	"UNDEF",
	"READ_REQUEST",
	"KEEP_ALIVE",
	"READ_HEADER",
	"READ_POST",
	"HANDLE_REQUEST",
	"HANDLING",
	"PREPARE_HEADER",
	"WRITE_HEADER",
	"WRITE_BODY",
	"WRITE_CONTENT",
	"WAIT_DATA",
#if SUPPORT_THROTTLE
	"THROTTLED",
#endif
	"DONE",
	"CLOSE",
	"NOT_OK"
};


#define REQUESTS_STEP 128

char* methods[] = {"?", "HEAD", "GET", "POST"};

FD* fds = NULL;
int maxFD = 0;
int fdLimit = 0;

time_t cleanupTime = 0;

char past[SHORT_STR];
char future[SHORT_STR];

int ruID = 0;

Request* currentRequest = NULL;

static Request* requests = NULL;
static Request* tail = NULL;
static Request* waiting = NULL;
static Request* waitingTail = NULL;
static Request* freeRequests = NULL;
static Request** requestsBlocks = NULL;
static unsigned nRequests = 0, maxRequests = 0, nRequestsBlocks = 0;

static unsigned totalAccepts = 0, totalRequests = 0;
static unsigned lastUser = 0, lastSys = 0, lastAccepts = 0, lastRequests = 0;

static int countsFD = -2;

#if MODEL == MODEL_RTSIG
int rtSignal = 0;
#elif MODEL == MODEL_KEVENT
int kQueue = 0;
int nChanges, aChanges;
struct kevent* kChanges;
intptr_t lastEventData = 0;
#elif MODEL == MODEL_EPOLL
int epoll = 0;
int aEvents = 0;
struct epoll_event *epollEvents;
unsigned lastEventData = 0;
int initEpoll();
#endif

char* getStatusLine(int code);
int makeErrorReply(Request* r, char* buffer, int size);


time_t
timerfc(char *s)
{
	static const int daytab[2][12] = {
		{0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 },
		{0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335 }
	};
	enum { D_START, D_END, D_MON, D_DAY, D_YEAR, D_HOUR, D_MIN, D_SEC};
	unsigned sec, min, hour, day, mon, year;
	char month[3];
	register char c;
	register unsigned n;
	register char flag;
	register char state;
	register char isctime;
	sec = 60;
	min = 60;
	hour = 24;
	day = 32;
	year = 1969;
	isctime = 0;
	month[0] = 0;
	state = D_START;
	n = 0;
	flag = 1;
	do {
		c = *s++;
		switch (state) {
		case D_START:
			if (c == ' ') {
				state = D_MON;
				isctime = 1;
			} else if (c == ',') state = D_DAY;
			break;
		case D_MON:
			if ('a' <= c && c <= 'z') {
				if (n < 3) month[n++] = c;
			} else {
				if (n < 3) return -1;
				n = 0;
				state = isctime ? D_DAY : D_YEAR;
			}
			break;
		case D_DAY:
			if (c == ' ' && flag)
				;
			else if ('0' <= c && c <= '9') {
				flag = 0;
				n = 10 * n + (c - '0');
			} else {
				day = n;
				n = 0;
				state = isctime ? D_HOUR : D_MON;
			}
			break;
		case D_YEAR:
			if ('0' <= c && c <= '9')
				n = 10 * n + (c - '0');
			else {
				year = n;
				n = 0;
				state = isctime ? D_END : D_HOUR;
			}
			break;
		case D_HOUR:
			if ('0' <= c && c <= '9')
				n = 10 * n + (c - '0');
			else {
				hour = n;
				n = 0;
				state = D_MIN;
			}
			break;
		case D_MIN:
			if ('0' <= c && c <= '9')
				n = 10 * n + (c - '0');
			else {
				min = n;
				n = 0;
				state = D_SEC;
			}
			break;
		case D_SEC:
			if ('0' <= c && c <= '9')
				n = 10 * n + (c - '0');
			else {
				sec = n;
				n = 0;
				state = isctime ? D_YEAR : D_END;
			}
			break;
		}
	} while (state != D_END && c);
	switch (month[0]) {
	case 'a':
		mon = (month[1] == 'p') ? 4 : 8;
		break;
	case 'd':
		mon = 12;
		break;
	case 'f':
		mon = 2;
		break;
	case 'j':
		mon = (month[1] == 'a') ? 1 : ((month[2] == 'l') ? 7 : 6);
		break;
	case 'm':
		mon = (month[2] == 'r') ? 3 : 5;
		break;
	case 'n':
		mon = 11;
		break;
	case 'o':
		mon = 10;
		break;
	case 's':
		mon = 9;
		break;
	default:
		return -1;
	}
	if (year <= 100)
		year += (year < 70) ? 2000 : 1900;
	--mon;
	--day;
	if (sec >= 60 || min >= 60 || hour >= 60 || day >= 31 || year < 1970)
		return -1;
	return sec + 60L * (min + 60L * (hour + 24L * (day + daytab[year % 4 == 0][mon] + 365L * (year - 1970L) + ((year - 1969L) >> 2))));
}



int
copyListener(Listener* listener, Listener* old)
{
	while (old) {
		if (listener->addr.s_addr == old->addr.s_addr && listener->port == old->port) {
			if (old->fd > 0) {
				listener->fd = old->fd;
				listener->accepts = old->accepts;
				listener->acceptsToday = old->acceptsToday;
				listener->acceptsLast = old->acceptsLast;
				listener->droppedS = old->droppedS;
				listener->droppedC = old->droppedC;

				fds[old->fd].listener = listener;
				old->fd = -1;
				return OK;
			}
			return FAILED;
		}
		old = old->next;
	}
	return FAILED;
}


int
initListeners(Listener* old)
{
	Listener* listener;
	int inited = 0;
	uid_t uid = getuid();
	uid_t euid = geteuid();

	for (listener = config->listeners; listener; listener = listener->next) {
		struct sockaddr_in sa;
		int fd;
		int onoff = 1;
		int err;

		debug("initListener(%s:%d)", inet_ntoa(listener->addr), listener->port);
		if (old && copyListener(listener, old)) {
			message(MESSAGE, "using old listener %s:%d", inet_ntoa(listener->addr), listener->port);
			inited++;
			continue;
		}

		if (uid != euid && seteuid(uid)) message(SYS|ERROR, "cannot seteuid(%d)", uid);
		fd = socket(AF_INET, SOCK_STREAM, 0);
		err = errno;
		if (uid != euid && seteuid(euid)) message(SYS|ERROR, "cannot seteuid(%d)", euid);

		if (fd < 0) {
			char e[LONG_STR];
			errno = err;
			snprintf(e, LONG_STR, "cannot create socket: %s", strerror(errno));
			listener->error = STRDUP(e);
			message(SYS|ERROR, "cannot create listener socket");
			continue;
		}

		if (!fd || fd >= fdLimit) {
			char e[LONG_STR];
			snprintf(e, LONG_STR, "bad fd: %d", fd);
			listener->error = STRDUP(e);
			message(ERROR, "cannot init listener %s:%d: bad fd (%d)", inet_ntoa(listener->addr), listener->port, fd);
			close(fd);
			continue;
		}

		if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &onoff, sizeof(onoff)))
			message(SYS|ERROR, "cannot setsockopt(SO_REUSEADDR)");

		//fcntl(fd, F_SETFD, FD_CLOEXEC);

#if MODEL == MODEL_RTSIG
		if (fcntl(fd, F_SETSIG, rtSignal))
			message(SYS|ERROR, "cannot fcntl(F_SETSIG)");
		if (fcntl(fd, F_SETOWN, myPID))
			message(SYS|ERROR, "cannot fcntl(F_SETOWN)");
		if (fcntl(fd, F_SETFL, O_NONBLOCK|O_ASYNC))
			message(SYS|ERROR, "cannot fcntl(F_SETFL)");
#elif MODEL == MODEL_KEVENT
		watchEvent(fd, EVFILT_READ, EV_ADD, 0, 0);
		if (fcntl(fd, F_SETFL, O_NONBLOCK))
			message(SYS|ERROR, "cannot fcntl(F_SETFL)");
#elif MODEL == MODEL_EPOLL
		watchEvent(fd, EPOLLIN, EPOLL_CTL_ADD);
		if (fcntl(fd, F_SETFL, O_NONBLOCK))
			message(SYS|ERROR, "cannot fcntl(F_SETFL)");
#else
		if (fcntl(fd, F_SETFL, O_NONBLOCK))
			message(SYS|ERROR, "cannot fcntl(F_SETFL)");
#endif

		memset(&sa, 0, sizeof(sa));
		sa.sin_family = AF_INET;
		sa.sin_addr = listener->addr;
		sa.sin_port = htons(listener->port);

		if (uid != euid && seteuid(uid)) message(SYS|ERROR, "cannot seteuid(%d)", uid);
		onoff = bind(fd, (struct sockaddr*) &sa, sizeof sa);
		err = errno;
		if (uid != euid && seteuid(euid)) message(SYS|ERROR, "cannot seteuid(%d)", euid);
		if (onoff < 0) {
			char e[LONG_STR];
			errno = err;
			snprintf(e, LONG_STR, "cannot bind: %s", strerror(errno));
			listener->error = STRDUP(e);
			message(SYS|ERROR, "cannot bind on %s:%d", inet_ntoa(listener->addr), listener->port);
			continue;
		}

		if (listen(fd, config->connections) < 0) {
			char e[LONG_STR];
			snprintf(e, LONG_STR, "cannot listen: %s", strerror(errno));
			listener->error = STRDUP(e);
			message(SYS|ERROR, "cannot listen %s:%d", inet_ntoa(listener->addr), listener->port);
			continue;
		}
		fds[fd].listener = listener;
		listener->fd = fd;

		message(MESSAGE, "listening %s:%d", inet_ntoa(listener->addr), listener->port);
		inited++;
	}
	if (!inited) return message(ERROR, "nothing to listen");

	{
		struct rlimit rl;
        char err[LONG_STR];
        *err = 0;
		if (uid != euid && seteuid(uid)) message(SYS|ERROR, "cannot seteuid(%d)", uid);
		getrlimit(RLIMIT_NOFILE, &rl);
		rl.rlim_cur = fdLimit;
		if (rl.rlim_max < (unsigned)fdLimit) rl.rlim_max = (unsigned)fdLimit;
		if (setrlimit(RLIMIT_NOFILE, &rl))
			strcat(err, fmt("cannot set RLIMIT_NOFILE=%d/%d: %s ",
							(int)(rl.rlim_cur), (int)(rl.rlim_max), strerror(errno)));

#if MODEL == MODEL_RTSIG
#ifdef RLIMIT_SIGPENDING
		getrlimit(RLIMIT_SIGPENDING, &rl);
		message(MESSAGE, "RLIMIT_SIGPENDING: %d/%d", (int)(rl.rlim_cur), (int)(rl.rlim_max));
		if (rl.rlim_cur < config->connections * 2) {
			rl.rlim_cur = rl.rlim_max = config->connections * 2;
			if (setrlimit(RLIMIT_SIGPENDING, &rl))
				strcat(err, fmt("cannot set RLIMIT_SIGPENDING=%d/%d: %s",
								(int)(rl.rlim_cur), (int)(rl.rlim_max), strerror(errno)));
		}
#endif
#endif

		if (uid != euid && seteuid(euid)) message(SYS|ERROR, "cannot seteuid(%d)", euid);
		if (*err) message(ERROR, "%s", err);
	}

	if (config->detachRoot) {
		uid_t gid = getgid();
		uid_t egid = getegid();

		if (uid != euid && setreuid(euid, euid)) message(SYS|ERROR, "cannot setreuid(%d, %d)", euid, euid);
		if (gid != egid && setregid(egid, egid)) message(SYS|ERROR, "cannot setregid(%d, %d)", euid, euid);
	}
	return OK;
}

char*
serverTokens()
{
	return
		config->serverTokens == 2 ? "Server: 0W" RN :
		config->serverTokens ? "Server: 0W/" VERSION RN :
		"";
}

int
makeBusyReply(char* buffer, int size, char* message)
{
	return catstr(buffer, size,
				  "HTTP/1.0 503 Service Unavailable" RN,
				  serverTokens(),
				  "Date: ", now_str, RN
				  "Content-type: text/plain" RN RN,
				  message ? : "The server is temporarily busy. Try again later." RN,
				  NULL);
}


#if SUPPORT_ANTIFLOOD
int
floodReply(char* reply, int size, IPInfo* i)
{
	return
		makeBusyReply(reply, size,
					  fmt("Too many requests from one IP: %s." RN
						  "Access denied for %s minutes." RN
						  "Limit: %s requests per %s minutes." RN
						  "Denial till:  %s" RN
						  "Current time: %s" RN,
						  ip2string(i->ip),
						  itos((config->antifloodExpired + 30)/60),
						  itos(config->requestsPerIP),
						  itos((config->antifloodPeriod + 30)/60),
						  time2str(i->stamp, HTTP_TIME),
						  now_str));
}

#endif


void
insertToRequests(Request* r)
{
	debug("insertToRequests(%s)", ip2string(r->ip));
	if ((r->next = requests)) requests->prev = r;
	else tail = r;
	requests = r;
	r->prev = NULL;
}

void
removeFromRequests(Request* r)
{
    debug("removeFromRequests(%s)", ip2string(r->ip));
	if (r->next) r->next->prev = r->prev;
	else tail = r->prev;
	if (r->prev) r->prev->next = r->next;
	else requests = r->next;
}

void
insertToWaiting(Request* r)
{
    debug("insertToWaiting(%s)", ip2string(r->ip));
	if ((r->next = waiting)) waiting->prev = r;
	else waitingTail = r;
	waiting = r;
	r->prev = NULL;
}

void
removeFromWaiting(Request* r)
{
    debug("removeFromWaiting(%s)", ip2string(r->ip));
	if (r->next) r->next->prev = r->prev;
	else waitingTail = r->prev;
	if (r->prev) r->prev->next = r->next;
	else waiting = r->next;
}

#if SUPPORT_THROTTLE
void
insertToThrottled(Server* server, Request* r)
{
	debug("insertToThrottled(%s, %s)", ip2string(r->ip), r->uri);
	if ((r->nextThrottled = server->throttled)) server->throttled->prevThrottled = r;
	else server->throttledTail = r;
	server->throttled = r;
	r->prevThrottled = NULL;
	r->throttled = YES;
}

void
removeFromThrottled(Server* server, Request* r)
{
    debug("removeFromThrottled(%s, %s)", ip2string(r->ip), r->uri);
	if (r->nextThrottled) r->nextThrottled->prevThrottled = r->prevThrottled;
	else server->throttledTail = r->prevThrottled;
	if (r->prevThrottled) r->prevThrottled->nextThrottled = r->nextThrottled;
	else server->throttled = r->nextThrottled;
    r->throttled = NO;
}

#endif /* SUPPORT_THROTTLE */

#if DEBUGLOG
void
showList()
{
    Request* r = requests;
	debug("Requests head: %p, tail: %p", requests, tail);
	while(r) {
		debug("r: %p, next=%p, prev=%p", r, r->next, r->prev);
        r = r->next;
	}
	debug("Waiting head: %p, tail: %p", waiting, waitingTail);
    r = waiting;
	while(r) {
		debug("r: %p, next=%p, prev=%p", r, r->next, r->prev);
        r = r->next;
	}
}
#endif



typedef struct Accepted_ {
	int fd;
	unsigned ip;
} Accepted;

void
acceptRequest(Listener* listener)
{
	Accepted accepts[2000];
#if MODEL == MODEL_KEVENT
    // number of accepts is reported by kernel
    int limit = min(2000, lastEventData);
#else
    int limit = 2000;
#endif
    Accepted* a = accepts;
	socklen_t lsa = sizeof(struct sockaddr_in);

	debug("%d> acceptRequest()", listener->fd);

	while (a < accepts + limit) {
		struct sockaddr_in sa;
		if ((a->fd = accept(listener->fd, (struct sockaddr *) &sa, &lsa)) < 0) {
			if (errno != EAGAIN && errno != ECONNABORTED) message(SYS|ERROR, "cannot accept");
			break;
		}
		a->ip = ntohl(sa.sin_addr.s_addr);
		debug("%d> accepted from %s", a->fd, ip2string(a->ip));
		a++;
	}

	if (a == accepts + 2000) message(MESSAGE, "warning: too high accepts rate");

	while (--a >= accepts) {
		Request* r;
#if SUPPORT_ACCESS
        char trusted;
#endif
#if SUPPORT_ANTIFLOOD
		IPInfo* ipInfo = NULL;
#endif

		totalAccepts++;
		listener->accepts++;
 		if (!a->fd || a->fd >= fdLimit) {
			char reply[SHORT_TEXT];
			int n = makeBusyReply(reply, SHORT_TEXT, NULL);
			fcntl(a->fd, F_SETFL, O_NONBLOCK);
			write(a->fd, reply, n);
			while (read(a->fd, reply, SHORT_TEXT) == SHORT_TEXT);
			if (close(a->fd)) message(SYS|ERROR, "cannot close");
			message(ERROR, "connection to %s dropped: bad fd (%d)", ip2string(a->ip), a->fd);
			listener->droppedS++;
			continue;
		}
		if (nRequests == config->connections) {
			char reply[SHORT_TEXT];
			int n = makeBusyReply(reply, SHORT_TEXT, NULL);
			fcntl(a->fd, F_SETFL, O_NONBLOCK);
			write(a->fd, reply, n);
			while (read(a->fd, reply, SHORT_TEXT) == SHORT_TEXT);
			if (close(a->fd)) message(SYS|ERROR, "cannot close");
			message(ERROR, "connection to %s dropped: too many requests", ip2string(a->ip));
			listener->droppedS++;
			continue;
		}

#if SUPPORT_ACCESS
		if (config->access) {
			Request req;
			Request* r = &req;
			memset(r, 0, sizeof(Request));
			r->ip = a->ip;
#if SUPPORT_COUNTRY
			r->country = -1;
#endif
			if ((r->status = checkAccess(r, config->access, &r->location))) {
				char reply[MEDIUM_TEXT];
				char loc[LONG_STR];
				int n;
				if (r->location) catstr(loc, LONG_STR, "Location: ", r->location, RN, NULL);
				else *loc = 0;
				fcntl(a->fd, F_SETFL, O_NONBLOCK);
				while (read(a->fd, reply, MEDIUM_TEXT) == MEDIUM_TEXT);

				n = catstr(reply, MEDIUM_TEXT,
						   "HTTP/1.0 ", itoa(r->status), " ",
						   getStatusLine(r->status), RN,
						   serverTokens(),
						   "Date: ", now_str, RN
						   "Content-type: text/html" RN,
						   loc, RN,
						   NULL);
				n += makeErrorReply(r, reply + n, MEDIUM_TEXT - n);
				write(a->fd, reply, n);
				if (close(a->fd)) message(SYS|ERROR, "cannot close");
				listener->droppedC++;
				continue;
			}
		}


		trusted = checkIP(a->ip, config->trusted);
		if (!trusted) {
#endif

#if SUPPORT_ANTIFLOOD
			if (!(ipInfo = addIPConnection(a->ip))) {
				char reply[SHORT_TEXT];
				int n = makeBusyReply(reply, SHORT_TEXT,
									  fmt("Too many simultaneous connections from one IP, current limit: %d." RN,
										  config->connectionsPerIP));
				fcntl(a->fd, F_SETFL, O_NONBLOCK);
				write(a->fd, reply, n);
				while (read(a->fd, reply, SHORT_TEXT) == SHORT_TEXT);
				if (close(a->fd)) message(SYS|ERROR, "cannot close");
				listener->droppedC++;
				debug("connection to %s dropped: too many connections per IP", ip2string(a->ip));
				continue;
			}

			if (ipInfo->denied) {
				char reply[SHORT_TEXT];
				int n = floodReply(reply, SHORT_TEXT, ipInfo);
				fcntl(a->fd, F_SETFL, O_NONBLOCK);
				write(a->fd, reply, n);
				while (read(a->fd, reply, SHORT_TEXT) == SHORT_TEXT);
				if (close(a->fd)) message(SYS|ERROR, "cannot close");
				listener->droppedC++;
                ipInfo->denied++;
				decIPConnection(ipInfo);
				debug("connection to %s dropped: too many requests per IP", ip2string(a->ip));
				continue;
			}
#endif

#if SUPPORT_ACCESS
		}
#endif
		if (!(r = freeRequests)) {
			debug("acceptRequest: allocating new requests block");
			freeRequests = CALLOC(REQUESTS_STEP * sizeof(Request));
			if (!freeRequests) {
				char reply[SHORT_TEXT];
				int n = makeBusyReply(reply, SHORT_TEXT, NULL);
				fcntl(a->fd, F_SETFL, O_NONBLOCK);
				write(a->fd, reply, n);
				while (read(a->fd, reply, SHORT_TEXT) == SHORT_TEXT);
				if (close(a->fd)) message(SYS|ERROR, "cannot close");
				message(ERROR, "connection to %s dropped: not enough memory", ip2string(a->ip));
#if SUPPORT_ANTIFLOOD
#if SUPPORT_ACCESS
				if (!trusted)
#endif
					decIPConnection(ipInfo);
#endif
				listener->droppedS++;
				continue;
			}
			for (r = freeRequests; r < freeRequests + REQUESTS_STEP - 1; r++)
				r->next = r + 1;
			r->next = NULL;
			r = requestsBlocks[nRequestsBlocks++] = freeRequests;
		}

		freeRequests = freeRequests->next;


		r->type = REQUEST;
		r->state = READ_REQUEST;
		r->fd = a->fd;
		r->rangeFrom = r->rangeTo = r->inContentLength =
			r->file = r->length = -1;
		r->major = 1;
		r->ip = a->ip;
#if SUPPORT_COUNTRY
		r->country = -1;
		r->region = -1;
#endif
#if SUPPORT_ACCESS
		r->trusted = trusted;
#endif

		r->start = now;
        r->last = now + config->keepAliveTimeout;
		insertToWaiting(r);


		fds[a->fd].request = r;

		if (a->fd > maxFD) maxFD = a->fd;

		if (++nRequests > maxRequests) maxRequests = nRequests;

#if MODEL == MODEL_RTSIG
		fcntl(a->fd, F_SETSIG, rtSignal);
		fcntl(a->fd, F_SETOWN, myPID);
		fcntl(a->fd, F_SETFL, O_NONBLOCK|O_ASYNC);
		http(r);
#elif MODEL == MODEL_KEVENT
		fcntl(a->fd, F_SETFL, O_NONBLOCK);
		watchEvent(a->fd, EVFILT_READ, EV_ADD, 0, 0);
		r->watchRead = YES;
#elif MODEL == MODEL_EPOLL
		fcntl(a->fd, F_SETFL, O_NONBLOCK);
		watchEvent(a->fd, EPOLLIN, EPOLL_CTL_ADD);
		r->watchRead = YES;
#else
		fcntl(a->fd, F_SETFL, O_NONBLOCK);
		r->watchRead = YES;
#endif
		//fcntl(a->fd, F_SETFD, FD_CLOEXEC);
	}
}


int
handleEvent(int fd)
{
	char* con_type;

	debug("%d> handleEvent()", fd);

	if (fd <= 0 || fd >= fdLimit)
		return message(MESSAGE, "bad fd: %d", fd);

	if ((con_type = fds[fd].type) == NULL) {
        debug("%d> event on unused fd", fd);
		return FAILED;
	}

	switch (*con_type) {
	case REQUEST: http(fds[fd].request); break;
	case LISTENER: acceptRequest(fds[fd].listener); break;
#if SUPPORT_PROXY
	case PROXY_REQUEST: proxy(fds[fd].proxy); break;
#endif
	}

	currentRequest = NULL;

	return OK;
}

void
getUsage()
{
	static int prevAccepts;
	static unsigned prevRequests;
	static struct rusage prevUsage;
	struct rusage usage;
	Server* server = config->servers;
	Listener* listener = config->listeners;
#if SUPPORT_PROXY
	Backend* backend = config->backends;
#endif

	getrusage(RUSAGE_SELF, &usage);

	lastUser = (usage.ru_utime.tv_sec - prevUsage.ru_utime.tv_sec)*1000L + (usage.ru_utime.tv_usec - prevUsage.ru_utime.tv_usec)/1000L;
	lastSys = (usage.ru_stime.tv_sec - prevUsage.ru_stime.tv_sec)*1000L + (usage.ru_stime.tv_usec - prevUsage.ru_stime.tv_usec)/1000L;
	lastAccepts = totalAccepts - prevAccepts;
	lastRequests = totalRequests - prevRequests;

	if (countsFD > 0) {
		char buf[NORM_STR];
		int n = catstr(buf, NORM_STR, itoa(totalAccepts), "\n", itoa(totalRequests), "\n", NULL);
		static int prev = 1111;
		write(countsFD, buf, n);
		if (prev > n) ftruncate(countsFD, n);
		prev = n;
		lseek(countsFD, 0, SEEK_SET);
	} else if (countsFD == -2) {
		if (strcmp(config->configName, "httpd.conf"))
			countsFD = -3;
		else if ((countsFD = open("data/httpd/counts", O_CREAT|O_WRONLY, 0644)) < 0)
			message(SYS|ERROR, "cannot open %s", "data/httpd/counts");
	}


	if (totalAccepts >= 2000000000) totalAccepts -= 2000000000;
	if (totalRequests >= 2000000000) totalRequests -= 2000000000;

	prevUsage = usage;
	prevAccepts = totalAccepts;
	prevRequests = totalRequests;

	while(server) {
#if SUPPORT_THROTTLE
		while (server->throttledTail && server->sent < server->sentLimit) {
			debug("throttling %s: can send %d bytes", server->name, server->sentLimit - server->sent);
			http(server->throttledTail);
		}
#endif /* SUPPORT_THROTTLE */


		server->requestsLast = server->requests;
		server->sentLast = server->sent;
		server->requestsToday += server->requests;
		server->sentToday += server->sent;

		server->requests = server->sent = 0;

		server = server->next;
	}

	while(listener) {
		listener->acceptsLast = listener->accepts;
		listener->acceptsToday += listener->accepts;

		listener->accepts = 0;
		listener = listener->next;
	}

#if SUPPORT_PROXY
	while (backend) {
		backend->requestsLast = backend->requests;
		backend->requestsToday += backend->requests;

		backend->requests = 0;
		backend = backend->next;
	}
#endif

}

void
resetToday()
{
	Listener* listener = config->listeners;
	Server* server = config->servers;
#if SUPPORT_PROXY
	Backend* backend = config->backends;
#endif
	int n = 0, sumCount = 0, sumDroppedS = 0, sumDroppedC = 0;
    long long sumSent = 0;

	maxRequests = 0;

	message(MESSAGE, "Day totals:");
	while(listener) {
		if (!listener->error)
			message(MESSAGE, "listener %s:%d: %s accepts, %s + %s dropped",
					listener->addr.s_addr ? inet_ntoa(listener->addr): "*",
					listener->port,
					itos(listener->acceptsToday),
					itos(listener->droppedS),
					itos(listener->droppedC));

		sumCount += listener->acceptsToday;
		sumDroppedS += listener->droppedS;
		sumDroppedC += listener->droppedC;
		n++;

		listener->acceptsToday = listener->droppedS = listener->droppedC = 0;
		listener = listener->next;
	}
	if (n > 1)
		message(MESSAGE, "%d listeners: %s accepts, %s + %s dropped",
                n, itos(sumCount), itos(sumDroppedS), itos(sumDroppedC));

    n = sumCount = 0;

	while(server) {
		message(MESSAGE, "server %s: %s requests, %sb sent",
				server->name,
				itos(server->requestsToday),
				humanSize(server->sentToday));

		sumCount += server->requestsToday;
		sumSent += server->sentToday;
		n++;

		server->requestsToday = 0;
		server->sentToday = 0;
		server = server->next;
	}
	if (n > 1)
		message(MESSAGE, "%d servers: %s requests, %sb sent",
				n, itos(sumCount), humanSize(sumSent));

	n = sumCount = 0;

	resetTodayFiles();

#if SUPPORT_PROXY
	while (backend) {
		message(MESSAGE, "backend %s:%d: %s requests",
				backend->name, backend->port,
				itos(backend->requestsToday));
		sumCount += backend->requestsToday;
		n++;

		backend->requestsToday = backend->maxRequests = 0;
		backend = backend->next;
	}
	if (n > 1)
		message(MESSAGE, "%d backends: %s requests",
				n, itos(sumCount));

	resetTodayProxy();
#endif

#if SUPPORT_ANTIFLOOD
	resetTodayFlood();
#endif

#if SUPPORT_CACHE
	resetTodayCache();
#endif
}

void
cleanup()
{
	static int prev_day;
	static time_t fiveSecTime = 0;
#if SUPPORT_COUNTRY
	static time_t rangesTime = 0;
#endif
#if SUPPORT_ANTIFLOOD
	static time_t ipTime = 0;
#endif
#if SUPPORT_CACHE
	static time_t cacheTime = 0;
#endif

	debug("cleanup");

#if SUPPORT_PROXY
	cleanupProxy();
#endif


	while(tail && now > tail->last) {
		debug("%d> timeout in %s", tail->fd, states[tail->state]);
		tail->state = CLOSE;
		http(tail);
	}
	while(waitingTail && now > waitingTail->last) {
		debug("%d> timeout in %s", waitingTail->fd, states[waitingTail->state]);
		waitingTail->state = CLOSE;
		http(waitingTail);
	}

	if (fiveSecTime/5 != now/5) {
		fiveSecTime = now;
		cleanupFiles();
		getUsage();
	}

#if SUPPORT_COUNTRY
	if (rangesTime/60 != now/60) {
		rangesTime = now;
		if (refreshRanges(countries)) ruID = code2id(countries, "ru");
		refreshRanges(regions);
	}
#endif

#if SUPPORT_ANTIFLOOD
	if (ipTime/16 != now/16) {
		ipTime = now;
		cleanupIP();
	}
#endif

#if SUPPORT_CACHE
	if (cacheTime/64 != now/64) {
		cacheTime = now;
		cleanupCache();
#if DEBUGLOG
        showPtrs(NO);
#endif
	}
#endif


	if (prev_day != now_day) {
		if (prev_day) resetToday();

		strcpy(past, time2str(now - 365*24*60*60, HTTP_TIME));
		strcpy(future, time2str(now + 365*24*60*60, HTTP_TIME));
		prev_day = now_day;
	}

	cleanupTime = now;
}



void
closeListeners(Listener* listener)
{
	while (listener) {
		if (listener->fd > 0) {
#if MODEL == MODEL_KEVENT
			unwatchEvent(listener->fd);
#endif
			if (close(listener->fd) < 0)
				message(SYS|ERROR, "cannot close %s:%d", inet_ntoa(listener->addr), listener->port);
			fds[listener->fd].listener = NULL;
			listener->fd = -1;
		}
		listener = listener->next;
	}
}

int
freeAll()
{
	unsigned i;

	message(MESSAGE, "freing all resources");

	closeListeners(config->listeners);

	while (waiting) {
		waiting->state = CLOSE;
		http(waiting);
	}
	while (requests) {
		requests->state = CLOSE;
		http(requests);
	}

	resetToday();
	fds = FREE(fds);
	for (i = 0; i < nRequestsBlocks; i++)
		FREE(requestsBlocks[i]);
	requestsBlocks = FREE(requestsBlocks);
	freeRequests = NULL;
	freeFiles();

#if SUPPORT_PROXY
	freeProxyRequests();
#endif


#if MODEL == MODEL_KEVENT
	kChanges = FREE(kChanges);
	nChanges = aChanges = 0;
#elif MODEL == MODEL_EPOLL
	if (epoll) close(epoll);
	epollEvents = FREE(epollEvents);
	epoll = aEvents = 0;
#endif

	nRequests = maxRequests = nRequestsBlocks = maxFD = 0;
	config = freeConfig(config);
	message(CLOSE_LOGS, NULL);

	if (countsFD > 0) close(countsFD);

    return OK;
}

int
copyServerInfo(Server* server, Server* old)
{
	while (old) {
		if (server->hash == old->hash && !strcmp(server->name, old->name)) {
			server->requests	  = old->requests;
			server->sent		  = old->sent;
			server->requestsLast  = old->requestsLast;
			server->sentLast	  = old->sentLast;
			server->requestsToday = old->requestsToday;
			server->sentToday	  = old->sentToday;
			return OK;
		}
		old = old->next;
	}
	return FAILED;
}

#if SUPPORT_PROXY
int
copyBackendInfo(Backend* backend, Backend* old)
{
	while (old) {
		if (backend->addr.s_addr == old->addr.s_addr && backend->port == old->port) {
			backend->requests	   = old->requests;
			backend->requestsToday = old->requestsToday;
			backend->maxRequests   = old->maxRequests;
			return OK;
		}
		old = old->next;
	}
	return FAILED;
}
#endif

int
reloadConfig(int argc, char** argv)
{
	Config* old = config;
	int oldLimit = fdLimit;
	Server* server;

	if (configured == now)
		return message(MESSAGE, "skipping double reload"), OK;

	message(MESSAGE, "reloading config");


	if (!readConfig(argc, argv)) {
		message(MESSAGE, "config error, using old config");
		config = old;
		return OK;
	}

	fdLimit = 2 * config->connections +
#if SUPPORT_PROXY
		config->proxyConnections +
#endif
		config->nOfListeners + 16;
	if (fdLimit > oldLimit &&
		!(fds = REALLOC(fds, sizeof(FD) * fdLimit)))
		return FAILED;

	if (config->connections > old->connections &&
		!(requestsBlocks = REALLOC(requestsBlocks, ((config->connections / REQUESTS_STEP) + 1) * sizeof(Request*))))
		return FAILED;

#if SUPPORT_PROXY
    closeProxyAliveRequests();

	if (config->proxyConnections > old->proxyConnections  &&
		!initProxyRequests())
		return FAILED;
#endif

	if (fdLimit > oldLimit && !initFiles())
		return FAILED;

	if (!initListeners(old->listeners))
		return FAILED;

	closeListeners(old->listeners);

	for (server = config->servers; server; server = server->next)
		copyServerInfo(server, old->servers);

#if SUPPORT_PROXY
	{
		Backend* backend;
		for (backend = config->backends; backend; backend = backend->next)
			copyBackendInfo(backend, old->backends);
	}
#endif

#if SUPPORT_CACHE
	{
		Pair* p;
        startListPairs(config->caches);
		while ((p = nextPair(config->caches))) {
            Cache* cache = p->value;
            Cache* oldCache = getPair(old->caches, cache->name);
			if (oldCache) {
				Request* r = requests;
				copyCache(cache, oldCache);
				while (r) {
					if (r->alias && r->alias->cache == oldCache)
						r->alias->cache = cache;
					r = r->next;
				}
			}
		}
	}
#endif

	if (old->active) {
		Server* server = old->servers;
		while(server) {
			Alias* alias = server->aliases;
			while(alias) {
				alias->keepalive = NO;
				alias = alias->next;
			}
#if SUPPORT_THROTTLE
			// cannot support throttling for unattached server,
			// just remove limit for active connections and
			// let already throttled connection to die on timeout
			server->sentLimit = INT_MAX;
#endif /* SUPPORT_THROTTLE */


			server = server->next;
		}
	} else {
		freeConfig(old);
	}

	if (nRequests > config->connections) {
		int i = nRequests - config->connections;
		Request* r = waiting;
		while (r && i--) {
			Request* next = r->next;
			r->state = CLOSE;
			http(r);
			r = next;
		}
        r = requests;
		while (r && i-- > 0) {
			Request* next = r->next;
			r->state = CLOSE;
			http(r);
			r = next;
		}
	}

#if SUPPORT_ANTIFLOOD
	{
		Request* r = waiting;
		while (r) {
			Request* next = r->next;
#if SUPPORT_ACCESS
			if (!r->trusted)
#endif
				if (getActive(r->ip) > config->connectionsPerIP) {
					r->state = CLOSE;
					http(r);
				}
			r = next;
		}
	}
#endif

	configured = now;

	return OK;
}

int
deactivate()
{
	Server* server = config->servers;
	while(server) {
		Alias* alias = server->aliases;
		while(alias) {
			alias->keepalive = NO;
			alias = alias->next;
		}
		server = server->next;
	}

#if SUPPORT_PROXY
    closeProxyAliveRequests();
#endif

	closeListeners(config->listeners);
    config->listeners = NULL;

	unlink(PIDS "/httpd");
	if (countsFD > 0) unlink("data/httpd/counts");

	message(MESSAGE, "deactivated");
	return OK;
}


int
init(int argc, char** argv)
{
	getTime();
	configured = now;
	strcpy(past, time2str(now - 365*24*60*60, HTTP_TIME));
	strcpy(future, time2str(now + 365*24*60*60, HTTP_TIME));

	return
		(readConfig(argc, argv) &&
		 (fdLimit = 2 * config->connections +
#if SUPPORT_PROXY
		  config->proxyConnections +
#endif
		  config->nOfListeners + 16) &&
		 (fds = CALLOC(sizeof(FD) * fdLimit)) &&
		 (requestsBlocks = CALLOC(((config->connections / REQUESTS_STEP) + 1) * sizeof(Request*))) &&
#if SUPPORT_PROXY
		 initProxyRequests() &&
#endif
		 initFiles() &&
#if MODEL == MODEL_EPOLL
         initEpoll() &&
#endif
		 initListeners(NULL));
}



#if MODEL == MODEL_POLL
int
fillListenersPoll(struct pollfd* pollfds)
{
	struct pollfd* pollfd = pollfds;
	Listener* listener = config->listeners;
	while (listener) {
		if (listener->fd > 0) {
			pollfd->events = POLLIN;
			pollfd->fd = listener->fd;
			pollfd++;
		}
		listener = listener->next;
	}
	return pollfd - pollfds;
}

int
fillPoll(struct pollfd* pollfds)
{
	Request* r;
	struct pollfd* pollfd = pollfds;

	for (r = waiting; r; r = r->next) {
		pollfd->events = POLLIN;
		pollfd->fd = r->fd;
		pollfd++;
	}
	for (r = requests; r; r = r->next) {
		short ev = r->watchRead ? POLLIN : r->watchWrite ? POLLOUT : 0;
		if (ev) {
			pollfd->events = ev;
			pollfd->fd = r->fd;
			pollfd++;
		}
	}
	return pollfd - pollfds;
}


char
work(int argc, char** argv)
{

	while (!doQuit) {
		struct pollfd* pollfds;
		int l;

		if (!init(argc, argv)) return FAILED;

		pollfds = CALLOC(sizeof(struct pollfd) * (config->nOfListeners + config->connections));
		l = fillListenersPoll(pollfds);

		while (!(doQuit | doRestart)) {
			int n = l;
			int i;
			int r;
			n += fillPoll(pollfds + n);
			n += fillProxyPoll(pollfds + n);
			debug("polling %d fds...", n);
			r = poll(pollfds, n, 5000);
			debug("ready %d fds", r);
			getTime();
			if (r > 0) {
				struct pollfd* pollfd = pollfds;
				if (r == 1) usleep(1000);
				for (i = 0; i < n && r; i++, pollfd++) {
					if (pollfd->revents) {
						handleEvent(pollfd->fd);
						r--;
					}
				}
			} else if (r < 0) {
				if (errno != EINTR) message(SYS|ERROR, "cannot poll");
				if (errno == ENOMEM) sleep(10);
			}
			if (now != cleanupTime) cleanup();
			if (doReload) {
				if (!reloadConfig(argc, argv)) {
					doQuit = YES;
				} else {
					FREE(pollfds);
					pollfds = CALLOC(sizeof(struct pollfd) * (config->nOfListeners + config->connections));
					l = fillListenersPoll(pollfds);
				}
				message(CLOSE_LOGS, NULL);
				doReload = NO;
			}
			if (doReExec && !doShutdown) {
                debug("forking working process");
				if (fork()) return freeAll();
                doReExec = NO;
				doShutdown = YES;
			}
			if (doShutdown) {
				if (config->listeners) deactivate();
				if (!(requests || waiting)) doQuit = YES;
			}
		}
		FREE(pollfds);
		freeAll();
		doRestart = NO;
	}
	return OK;
}

#elif MODEL == MODEL_RTSIG
char
work(int argc, char** argv)
{
	//	int count = 0;
	sigset_t sigset;
	rtSignal = SIGRTMIN;
	sigemptyset(&sigset);
	sigaddset(&sigset, rtSignal);
	sigprocmask(SIG_BLOCK, &sigset, NULL);

	while (!doQuit) {
		if (!init(argc, argv)) return FAILED;
		while (!(doQuit | doRestart)) {
			int r;
			siginfo_t info;
			sigset_t set = sigset;
			struct timespec timeout = {5, 0};
			memset(&info, 0, sizeof(siginfo_t));

			debug("waiting for signal...");
			r = sigtimedwait(&set, &info, &timeout);
			//if ((count++ & 0xFF) == 0) getTime();
			getTime();
			if (r > 0)
				handleEvent(info.si_fd);
			else if (errno != EINTR && errno != EAGAIN)
				message(SYS|ERROR, "cannot sigwaitinfo");
			if (now != cleanupTime) cleanup();
			if (doReload) {
				if (!reloadConfig(argc, argv)) doQuit = YES;
				message(CLOSE_LOGS, NULL);
				doReload = NO;
			}
			if (doReExec && !doShutdown) {
                debug("forking working process");
				if (fork()) return freeAll();
                doReExec = NO;
				doShutdown = YES;
			}
			if (doShutdown) {
				if (config->listeners) deactivate();
				if (!(requests || waiting)) doQuit = YES;
			}
		}
		freeAll();
		doRestart = NO;
	}
	return OK;
}

#elif MODEL == MODEL_KEVENT
char*
showKeventFilter(short filter)
{
	static char str[MEDIUM_STR];

	strcpy(str,
		   (filter == EVFILT_READ) ? "EVFILT_READ" :
		   (filter == EVFILT_WRITE) ? "EVFILT_WRITE" :
		   (filter == EVFILT_AIO) ? "EVFILT_AIO" :
		   (filter == EVFILT_VNODE) ? "EVFILT_VNODE" :
		   (filter == EVFILT_PROC) ? "EVFILT_PROC" :
		   (filter == EVFILT_SIGNAL) ? "EVFILT_SIGNAL" :
		   (filter == EVFILT_TIMER) ? "EVFILT_TIMER" :
		   (filter == EVFILT_NETDEV) ? "EVFILT_NETDEV" : "???");

	return str;
}

char*
showKeventFlags(unsigned flags)
{
	static char str[MEDIUM_STR];
	int n =
		catstr(str, MEDIUM_STR,
			   (flags & EV_ADD) ? "EV_ADD|" : "",
			   (flags & EV_ENABLE) ? "EV_ENABLE|" : "",
			   (flags & EV_DISABLE) ? "EV_DISABLE|" : "",
			   (flags & EV_DELETE) ? "EV_DELETE|" : "",
			   (flags & EV_ONESHOT) ? "EV_ONESHOT|" : "",
			   (flags & EV_CLEAR) ? "EV_CLEAR|" : "",
			   (flags & EV_EOF) ? "EV_EOF|" : "",
			   (flags & EV_ERROR) ? "EV_ERROR|" : "",
			   NULL);
	if (n) str[--n] = 0;

	return str;
}

void
watchEvent(int fd, int filter, int flags, u_int fflags, int data)
{
	if (nChanges == aChanges) {
		if (aChanges) aChanges <<= 1;
		else aChanges = 128;
		kChanges = REALLOC(kChanges, aChanges * sizeof(struct kevent));
	}

	if (fd <= 0) {
		message(MESSAGE, "bad fd in watchEvent(%d, %s, %s, %d, %d)", fd,
				showKeventFilter(filter), showKeventFlags(flags), fflags, data);
		return;
	}

	debug("%d> watching event[%d]: %s %s (%d,%d)", fd, nChanges,
          showKeventFilter(filter), showKeventFlags(flags), fflags, data);
	EV_SET(kChanges + nChanges, fd, filter, flags, fflags, data, NULL);
	nChanges++;
}

void
unwatchEvent(int fd)
{
	int i;
    struct kevent *ke = kChanges;
	for (i = 0; i < nChanges; i++) {
		if (ke->ident == fd) {
			debug("%d> removing from new events at %d/%d", fd, i, nChanges);
			if (i < --nChanges) memmove(ke, ke + 1, (nChanges - i) * sizeof(struct kevent));
			--i;
		}
	}
}

char
work(int argc, char** argv)
{
	if ((kQueue = kqueue()) < 0)
		return message(SYS|ERROR, "cannot create kqueue");

	while (!doQuit) {
		if (!init(argc, argv)) return FAILED;
		while (!(doQuit | doRestart)) {
			struct timespec timeout = {5, 0};
			int r;
			struct kevent newEvents[4096];


			debug("waiting for event (changed: %d)...", nChanges);
			r = kevent(kQueue, kChanges, nChanges,
					   newEvents, 4096,
					   &timeout);
			nChanges = 0;
			getTime();
			if (r > 0) {
				struct kevent* ke = newEvents;
				int i;
				if (r == 1) usleep(1000);
				for (i = 0; i < r; i++, ke++) {
					debug("%d> event[%d]: %s %s (%d, %d)",
						  ke->ident, i, showKeventFilter(ke->filter),
						  showKeventFlags(ke->flags), ke->fflags, ke->data);
					lastEventData = ke->data;
					handleEvent(ke->ident);
				}
			} else if (r < 0 && errno != EINTR && errno != EAGAIN) {
				message(SYS|ERROR, "cannot wait event");
			}
			if (now != cleanupTime) cleanup();
			if (doReload) {
				if (!reloadConfig(argc, argv)) doQuit = YES;
				message(CLOSE_LOGS, NULL);
				doReload = NO;
			}
			if (doReExec && !doShutdown) {
				debug("forking working process");
				if (fork()) return freeAll();
                doReExec = NO;
				doShutdown = YES;
			}
			if (doShutdown) {
				if (config->listeners) deactivate();
				if (!(requests || waiting)) doQuit = YES;
			}

		}
		freeAll();
		doRestart = NO;
	}
	return OK;
}

#elif MODEL == MODEL_EPOLL

char*
showEpollFlags(unsigned flags)
{
	static char str[MEDIUM_STR];
	int n =
		catstr(str, MEDIUM_STR,
			   (flags & EPOLLIN) ? "EPOLLIN|" : "",
			   (flags & EPOLLOUT) ? "EPOLLOUT|" : "",
			   (flags & EPOLLPRI) ? "EPOLLPRI|" : "",
			   (flags & EPOLLRDNORM) ? "EPOLLRDNORM|" : "",
			   (flags & EPOLLRDBAND) ? "EPOLLRDBAND|" : "",
			   (flags & EPOLLWRNORM) ? "EPOLLWRNORM|" : "",
			   (flags & EPOLLWRBAND) ? "EPOLLWRBAND|" : "",
			   (flags & EPOLLMSG) ? "EPOLLMSG|" : "",
			   (flags & EPOLLERR) ? "EPOLLERR|" : "",
			   (flags & EPOLLHUP) ? "EPOLLHUP|" : "",
			   (flags & EPOLLONESHOT) ? "EPOLLONESHOT|" : "",
			   (flags & EPOLLET) ? "EPOLLET|" : "",
			   NULL);
	if (n) str[--n] = 0;

	return str;
}

int
initEpoll()
{
	if ((epoll = epoll_create(config->connections / 2)) < 0)
		return message(SYS|ERROR, "cannot create epoll");

	aEvents = 128;
	if (!(epollEvents = MALLOC(aEvents * sizeof(struct epoll_event))))
		return FAILED;
	return OK;
}

void
watchEvent(int fd, int events, int op)
{
    struct epoll_event ev;
	debug("Watching event: fd=%d, %s %s", fd, showEpollFlags(events),
		  op == EPOLL_CTL_ADD ? "add" : op == EPOLL_CTL_DEL ? "del" : "mod");

	if (fd <= 0) {
        message(MESSAGE, "bad fd in watchEvent(%d, %s %s)", fd, showEpollFlags(events),
				op == EPOLL_CTL_ADD ? "add" : op == EPOLL_CTL_DEL ? "del" : "mod");
		return;
	}

	ev.events = events /*| EPOLLET*/;
	ev.data.fd = fd;

	if (epoll_ctl(epoll, op, fd, &ev) < 0)
		message(SYS|ERROR, "cannot epoll_ctl(%d, %s, %s)", fd,
				op == EPOLL_CTL_ADD ? "add" : op == EPOLL_CTL_DEL ? "delete" : "mod",
				showEpollFlags(events));
}

char
work(int argc, char** argv)
{
	while (!doQuit) {
		if (!init(argc, argv)) return FAILED;

		while (!(doQuit | doRestart)) {
			int timeout = 5000;
			int r;

			debug("waiting for event...");
			r = epoll_wait(epoll, epollEvents, aEvents, timeout);

			if (r == aEvents) {
				aEvents <<= 1;
				debug("increasing events array to %d", aEvents);
				epollEvents = REALLOC(epollEvents, aEvents * sizeof(struct epoll_event));
			}

			getTime();
			if (r > 0) {
				struct epoll_event* ev = epollEvents;
				int i;
				if (r == 1) usleep(1000);
				for (i = 0; i < r; i++, ev++) {
					debug("Event: fd=%d, events=%s", ev->data.fd, showEpollFlags(ev->events));
					lastEventData = ev->events;
					handleEvent(ev->data.fd);
				}
			} else if (r < 0 && errno != EINTR && errno != EAGAIN) {
				message(SYS|ERROR, "cannot wait event");
			}
			if (now != cleanupTime) cleanup();
			if (doReload) {
				if (!reloadConfig(argc, argv)) doQuit = YES;
				message(CLOSE_LOGS, NULL);
				doReload = NO;
			}
			if (doReExec && !doShutdown) {
				debug("forking working process");
				if (fork()) return freeAll();
                doReExec = NO;
				doShutdown = YES;
			}
			if (doShutdown) {
				if (config->listeners) deactivate();
				if (!(requests || waiting)) doQuit = YES;
			}

		}
		freeAll();
		doRestart = NO;
	}
	return OK;
}

#else
#error "select model not ready yet"
#endif /* MODEL */


int
parseRequestLine(char* s, Request* r)
{
	char* e;

	if (*(int*)s == 0x20544547) { //"GET"
		r->method = GET;
		s += 4;
	} else if (*(int*)s == 0x54534f50 && s[4] == ' ') { //"POST "
		r->method = POST;
		s += 5;
	} else if (*(int*)s == 0x44414548 && s[4] == ' ') { //"HEAD "
		r->method = HEAD;
		s += 5;
	} else {
		return message(MESSAGE, "501-bad method: %s", s), 501;
	}

	r->uri = s;
	if (!(s = strchr(r->uri, ' '))) return message(MESSAGE, "400-no proto: %s", r->uri), 400;
	if (s - r->uri >= MAX_URI) return 414;

	*s++ = 0;
	if (*(int*)s != 0x50545448 || s[4] != '/') //"HTTP/"
		return message(MESSAGE, "400-bad proto: %s", s), 400;
	s += 5;

	r->major = strtol(s, &e, 10);
	if (e == s) return message(MESSAGE, "400-bad major: %s", s), 400;
	if (r->major < 1) r->major = 1;
	else if (r->major > 1) return message(MESSAGE, "505-bad major: %d", r->major), 505;

	s = e;
	if (*s++ != '.') return message(MESSAGE, "400-no dot proto: %s", s), 400;
	r->minor = strtol(s, &e, 10);
	if (e == s) return message(MESSAGE, "400-bad minor: %s", s), 400;
	if (r->minor < 0) r->minor = 0;

	r->keepalive = r->minor;

	return 0;
}

char*
unifyHostName(char* host)
{
	char* port;

	strtolower(host);
	if (config->skipWWW && *(int*)host == 0x2E777777) // "www."
		host += strlen("www.");

	if ((port = strchr(host, ':'))) {
		if (config->skipPort) *port = 0;
		else if (!strcmp(port, ":80")) *port = 0;
	}

	port = host + strlen(host) - 1;
	while (*port == '.' && port >= host) *port-- = 0;

	return host;
}


int
parseHeaderLine(char* line, Request* r)
{
	char* handled = NULL;
	char* value = strchr(line, ':');
	if (!value) return message(MESSAGE, "strange header line: %s", line);
	*value++ = 0;
	while (isSpace(*value)) value++;
	if (!*value) return 0;

	strtolower(line);

	switch (*line++) {
	case 'a':
		if (!strcmp(line, "uthorization")) r->authorization = handled = value;
#if SUPPORT_LANGUAGES
		else if (!strcmp(line, "ccept-language")) r->acceptLanguage = handled = value;
#endif
		break;

	case 'c':
		if (!strcmp(line, "ookie")) r->cookie = handled = value;
		else if (!strcmp(line, "onnection")) {
			strtolower(handled = value);
			if (r->minor) {
				if (!strcmp(value, "close"))
					r->keepalive = NO;
			} else {
				if (!strcasecmp(value, "keep-alive"))
					r->keepalive = YES;
			}
#if SUPPORT_CACHE
		} else if (!strcmp(line, "ache-control")) {
			if (strstr(value, "no-cache")) {
#if SUPPORT_ACCESS
                // force refreshing by trusted user
				r->nocache = r->trusted ? -1 : YES;
#else
				r->nocache = YES;
#endif
			} else if (strstr(value, "no-store")) {
#if SUPPORT_ACCESS
				// force deleting cache entry by trusted user
				r->nocache = r->trusted ? -2 : YES;
#else
				r->nocache = YES;
#endif
			}
#endif
		} else if (r->method == POST) {
			if (!strcmp(line, "ontent-type")) r->inContentType = handled = value;
			else if (!strcmp(line, "ontent-length")) r->inContentLength = atoi(handled = value);
		}
		break;

#if 0
	case 'f':
		if (!strcmp(line, "orwarded")) {
			if ((value = rindex(handled = value, ' '))) {
				r->forwardedFor = string2ip(value + 1);
				if (r->forwardedFor < 0x01000000) r->forwardedFor = -1;
				//message(MESSAGE, "strange ff2: %s\n", value + 1);
			}
		}
		break;
#endif

	case 'h':
		if (!strcmp(line, "ost")) r->host = unifyHostName(handled = value);
		break;

	case 'i':
		if (!strcmp(line, "f-modified-since")) {
			r->ims = timerfc(strtolower(handled = value));
			if (r->ims == (time_t) -1) r->ims = 0;
		} else if (!strcmp(line, "f-unmodified-since")) {
			r->iums = timerfc(strtolower(handled = value));
			if (r->iums == (time_t) -1) r->iums = 0;
		} else if (!strcmp(line, "f-range")) {
			r->ifRange = timerfc(strtolower(handled = value));
			if (r->ifRange == (time_t) -1) r->ifRange = 0;
		}
		break;

	case 'k':
		if (!strcmp(line, "eep-alive")) handled = value;
		break;

#if SUPPORT_CACHE
	case 'p':
		if (!strcmp(line, "ragma") && strstr(value, "no-cache")) r->nocache = YES;
		break;
#endif

	case 'r':
		if (!strcmp(line, "eferer")) r->referrer = handled = value;
		else if (!strcmp(line, "ange")) {
			ssize_t n;
			char *end;
			if (strncmp(handled = value, "bytes=", strlen("bytes="))) return 0;
			value += strlen("bytes=");
			errno = 0;
			if (*value == '-') { // last N bytes: bytes=-N
				n = strtol(++value, &end, 10);
				if (n && end > value && !*end && !errno) r->rangeTo = n;
				return 0;
			}


			n = strtol(value, &end, 10);
			if (end == value || *end != '-' || errno) return 0;
			r->rangeFrom = n;

			value = end + 1;
			if (!*value) return 0; // skip first N bytes: bytes=N-

			errno = 0;
			n = strtol(value, &end, 10);
			if (end == value || *end != 0 || errno) return 0;
			r->rangeTo = n;		   // bytes=M-N
		}
		break;

	case 'u':
		if (!strcmp(line, "ser-agent")) r->userAgent = handled = value;
		break;

	case 'x':
		if (!strcmp(line, "-real-ip")) {
			r->realIP = string2ip(handled = value);
#if SUPPORT_ACCESS
			if (r->trusted) {
				r->ip = r->realIP;
#if SUPPORT_COUNTRY
				r->country = ip2id(countries, r->ip);
				r->region = ip2id(regions, r->ip);
#endif
			}
#endif
		} else if (!strcmp(line, "-country")) {
		} else if (!strcmp(line, "-region")) {
		} else if (!strcmp(line, "-forwarded-for")) {
			r->forwardedFor = string2ip(handled = value);
			if (r->forwardedFor < 0x01000000) r->forwardedFor = -1;
			//message(MESSAGE, "strange ff1: %s\n", value);
		}
		break;
	}

#if MAX_OTHER_HEADERS
	if (!handled && r->inHeadersCount < MAX_OTHER_HEADERS) {
		--line;
		line[strlen(line)] = ':';
		r->inHeaders[r->inHeadersCount++] = line;
	}
#endif
	return 0;
}

#define COPY_FIELD(name) if (r->name >= start && r->name < end) r->name += diff

int
increaseRequestBuffer(Request* r)
{
	off_t diff;
	Buffer* src = r->buffer;
	char* start = src->buffer;
	char* end = src->buffer + src->count;
	if (!(r->buffer = newBuffer(src->size << 1))) {
		r->buffer = src;
		return FAILED;
	}
	memcpy(r->buffer->buffer, src->buffer,
		   r->buffer->count = src->count);
	r->buffer->start = src->start;

	diff = r->buffer->buffer - src->buffer;
	COPY_FIELD(uri);
	COPY_FIELD(path);
	COPY_FIELD(params);
	COPY_FIELD(query);
	COPY_FIELD(userAgent);
	COPY_FIELD(referrer);
	COPY_FIELD(authorization);
	COPY_FIELD(user);
	COPY_FIELD(cookie);
	COPY_FIELD(host);
	COPY_FIELD(inContentType);
#if SUPPORT_LANGUAGES
	COPY_FIELD(acceptLanguage);
#endif
#if MAX_OTHER_HEADERS
	{
		int i;
		for (i = 0; i < r->inHeadersCount; i++) COPY_FIELD(inHeaders[i]);
	}
#endif

	COPY_FIELD(contentType);
	COPY_FIELD(contentEnc);
	COPY_FIELD(location);
	COPY_FIELD(auth);
	COPY_FIELD(setCookie);
	COPY_FIELD(body);
#if MAX_OTHER_HEADERS
	{
		int i;
		for (i = 0; i < r->outHeadersCount; i++) COPY_FIELD(outHeaders[i]);
	}
#endif
	freeBuffer(src);
	return OK;
}

char*
addStringToBuffer(Request* r, char* string, int n)
{
	char* s;

	if (!r->buffer) return NULL;
	while (r->buffer->count + n > r->buffer->size) {
		debug("addStringToBuffer(%s): doubling buffer", string);
		if (!increaseRequestBuffer(r)) return NULL;
	}

	s = r->buffer->buffer + r->buffer->start;
	if ((r->buffer->count - r->buffer->start) > 0) // pipelined
		memmove(s + n, s, r->buffer->count - r->buffer->start);

	memcpy(s, string, n);
	r->buffer->count += n;
	r->buffer->start += n;

	return s;
}



char*
getLine(Request* r)
{
	if (!r || r->fd <= 0) {
		message(ERROR, "getLine for bad request: %s", r ? "not connected" : "NULL");
		return NULL;
	}

	while (1) {
		ssize_t cnt;
		if (r->buffer->count - r->buffer->start > 0) {
			char* line = r->buffer->buffer + r->buffer->start;
			char* eol;
			if ((eol = memchr(line, '\n', r->buffer->count - r->buffer->start))) {
				r->buffer->start = eol + 1 - r->buffer->buffer;
				*eol-- = 0;
				while (eol >= line && isSpace(*eol)) *eol-- = 0;
				return line;
			}
		}

		if (r->buffer->count == r->buffer->size)  {
			if (r->buffer->size < MAXIMUM_BUFFER_SIZE) {
				debug("getLine(): doubling buffer");
				if (!increaseRequestBuffer(r)) return NULL;
			} else {
				debug("buffer overflow while reading from %s", ip2string(r->ip));
				return NULL;
			}
		}

		errno = 0;
		cnt = read(r->fd, r->buffer->buffer + r->buffer->count, r->buffer->size - r->buffer->count);
		debug("%d> reading %d bytes (%s)", r->fd, cnt, errno?strerror(errno):"OK");
		if (cnt <= 0) {
			if (errno == EWOULDBLOCK) return NULL;

			if (cnt < 0 && errno != ECONNRESET) message(SYS|ERROR, "cannot read from %s", ip2string(r->ip));
			return NULL;
		}
		r->buffer->count += cnt;
	}
}

char*
getStatusLine(int code)
{
	switch (code) {
	case 100: return "Continue";
	case 101: return "Switching Protocols";

	case 200: return "OK";
	case 201: return "Created";
	case 202: return "Accepted";
	case 203: return "Non-Authoritative Information";
	case 204: return "No Content";
	case 205: return "Reset Content";
	case 206: return "Partial Content";

	case 300: return "Multiple Choices";
	case 301: return "Moved Permanently";
	case 302: return "Moved Temporarily";
	case 303: return "See Other";
	case 304: return "Not Modified";
	case 305: return "Use Proxy";

	case 400: return "Bad Request";
	case 401: return "Unauthorized";
	case 402: return "Payment Required";
	case 403: return "Forbidden";
	case 404: return "Not Found";
	case 405: return "Method Not Allowed";
	case 406: return "Not Acceptable";
	case 407: return "Proxy Authentication Required";
	case 408: return "Request Timeout";
	case 409: return "Conflict";
	case 410: return "Gone";
	case 411: return "Length Required";
	case 412: return "Precondition Failed";
	case 413: return "Request Entity Too Large";
	case 414: return "Request-URI Too Long";
	case 415: return "Unsupported Media Type";
	case 416: return "Requested Range Not Satisfiable";

	case 500: return "Internal Server Error";
	case 501: return "Not Implemented";
	case 502: return "Bad Gateway";
	case 503: return "Service Unavailable";
	case 504: return "Gateway Timeout";
	case 505: return "HTTP Version Not Supported";
	default:  return "";
	}
}

int
fillHeaders(Request* r, char* buffer, int size)
{
	int l = 0;
	char* status = getStatusLine(r->status);
	char* b = buffer;

	debug("%d> fillHeaders()", r->fd);
	size -= 2;

	l = catstr(b, size,
			   "HTTP/", itoa(r->major), ".", itoa(r->minor), " ",
			   itoa(r->status), " ", status, RN,
			   "Date: ", now_str, RN,
			   serverTokens(),
			   NULL);

	if (r->keepalive) {
		if (r->minor == 0)
			l += catstr(b + l, size - l, "Connection: Keep-Alive" RN, NULL);
	} else if (r->minor) {
		l += catstr(b + l, size - l, "Connection: Close" RN, NULL);
	}

	if (r->status == 405)
		l += catstr(b + l, size - l, "Allow: GET, HEAD, POST" RN, NULL);

	if (r->contentType)
		l += catstr(b + l, size - l, "Content-Type: ", r->contentType, RN, NULL);

	if (r->contentEnc)
		l += catstr(b + l, size - l, "Content-Encoding: ", r->contentEnc, RN, NULL);

	if (r->location)
		l += catstr(b + l, size - l, "Location: ", r->location, RN, NULL);

	if (r->rangeTotal)
		l += catstr(b + l, size - l, "Content-Range: bytes ",
					itoa(r->rangeFrom), "-",
					itoa(r->rangeTo), "/",
					itoa(r->rangeTotal), RN,
					NULL);

	if (r->length >= 0)
		l += catstr(b + l, size - l, "Content-Length: ", itoa(r->length), RN, NULL);

	if (r->acceptRanges)
		l += catstr(b + l, size - l, "Accept-Ranges: bytes" RN, NULL);

	if (r->modified > 0)
		l += catstr(b + l, size - l, "Last-Modified: ", time2str(r->modified, HTTP_TIME), RN, NULL);

	if (r->auth)
		l += catstr(b + l, size - l, "WWW-Authenticate: ", r->auth, RN, NULL);

	if (r->status >= 500 && r->expires == EXPIRES_FUTURE)
		r->expires = EXPIRES_PAST;

	if (r->expires == EXPIRES_PAST)
		l += catstr(b + l, size - l,
					"Expires: ", past, RN
					"Pragma: no-cache" RN
					"Cache-control: no-cache" RN,
					NULL);
	else if (r->expires == EXPIRES_FUTURE)
		l += catstr(b + l, size - l, "Expires: ", future, RN, NULL);
	else if (r->expires)
		l += catstr(b + l, size - l, "Expires: ", time2str(r->expires, HTTP_TIME), RN, NULL);

	if (r->setCookie)
		l += catstr(b + l, size - l,
					"P3P: policyref=\"/w3c/p3p.xml\", CP=\"UNI\"" RN
					"Set-Cookie: ", r->setCookie, RN,
					NULL);

#if MAX_OTHER_HEADERS
	{
		int i;
		for (i = 0; i < r->outHeadersCount; i++) {
			int n = strlen(r->outHeaders[i]);
			if (l + n + 2 < size) {
				memcpy(b + l, r->outHeaders[i], n); l += n;
				memcpy(b + l, RN, 2); l += 2;
			}
		}
	}
#endif

	memcpy(b + l, RN, 2);
	return l + 2;
}

int
makeErrorReply(Request* r, char* buffer, int size)
{
	int l;
	char* status = getStatusLine(r->status);
	char* b = buffer;
    ErrorPages* e;

	debug("%d> makeErrorReply()", r->fd);

	while ((r->server && (e = r->server->errorPages)) || (e = config->errorPages)) {
		char* filename;
        int fd;
		if (e->codes) {
			int i = 0;
			do {
				if (e->codes[i++] == r->status) goto foundCode;
			} while (e->codes[i]);
		} else if ((400 <= r->status && r->status <= 409) ||
				   (500 <= r->status && r->status <= 509)) {
			goto foundCode;
		}

		break;

	foundCode:
		filename = fmt(e->format, r->status);
		if ((fd = getFD(filename)) < 0) {
			message(SYS|ERROR, "cannot open %s", filename);
            break;
		} else {
			lseek(fd, 0, SEEK_SET);
			l = read(fd, buffer, size);
			freeFD(fd);

			if (e == config->errorPages) {
				r->contentType = getPair(config->extensions, "html") ? : "text/html";
			} else {
				r->contentType =
					getPair(r->server->extensions, "html") ? :
					getPair(config->extensions, "html") ? :
					"text/html";
			}

			return l;
		}
	}

	l = catstr(b, size, "<html><head><title>[", itoa(r->status) ,"] ", status,
			   "</title></head>\n<body><h3>", status, "</h3>\n", NULL);
	switch (r->status) {
	case 301:
	case 302:
		l += catstr(b + l, size - l, "This document has moved <a href=\"",
					escapeLink(r->location), "\">here</a>.", NULL);
		break;
	case 401:
		l += catstr(b + l, size - l, "You need proper authorization to use this resource.", NULL);
		break;
	case 408:
		l += catstr(b + l, size - l, "Time is out for waiting your request.", NULL);
		break;
	case 400:
	case 405:
	case 501:
	case 505:
		l += catstr(b + l, size - l, "Your request was not understood or not allowed by this server.", NULL);
		break;
	case 403:
		l += catstr(b + l, size - l, "Access to this resource has been denied to you.", NULL);
		break;
	case 404:
		if (r->server)
			l += catstr(b + l, size - l, "Requested file <b>", quote(r->uri),
						"</b> could not be found on server <b>", r->host ? : r->server->name,
						"</b>.", NULL);
		else
			l += catstr(b + l, size - l, "Requested virtual server <b>",
						quote(r->host), "</b> could not be found.", NULL);
		break;
	case 411:
		l += catstr(b + l, size - l, "Missing Content-Length in post request.", NULL);
		break;
	case 412:
		l += catstr(b + l, size - l, "The precondition given in the request evaluated to false.", NULL);
		break;
	case 413:
		l += catstr(b + l, size - l, "The post request is too large.", NULL);
		break;
	case 416:
		l += catstr(b + l, size - l, "The requested range does not overlap with requested file.", NULL);
		break;
	case 503:
		l += catstr(b + l, size - l, "The server is temporarily busy. Try again later.", NULL);
		break;
	case 504:
		l += catstr(b + l, size - l, "The upstream server does not response. Try again later.", NULL);
		break;
	default:
		l += catstr(b + l, size - l, "An internal server error has occurred.", NULL);
		break;
	}
	if (r->server && r->server->admin)
		l += catstr(b + l, size - l, "\n<hr>If you beleive this is incorrect contact the site administrator at "
					"<a href=\"mailto:", r->server->admin, "\">",
					r->server->admin, "</a>.",
					NULL);

	if (l < 512 && r->status >= 400 && r->status != 503 && r->status != 504 &&
		r->userAgent && strstr(r->userAgent, "MSIE"))
		l += catstr(b + l, size - l,
					"\n"
					"<!--                                                                 \n"
					"    MSIE workaround: making page of 512+ bytes.                      \n"
					"                                                                     \n"
					"    This dummy text prevents MSIE from showing its own               \n"
					"    friendly error page. Read about this MSIE feature on the page    \n"
					"    http://support.microsoft.com/default.aspx?scid=kb;EN-US;218155   \n"
					"-->\n", NULL);


	l += catstr(b + l, size - l, "\n</body></html>\n", NULL);
	return l;
}

void
logRequest(Request* r)
{
	char* methods[] = {"-", "HEAD", "GET", "POST"};
	char* strings[] = {r->cookie, r->uri, r->referrer, r->userAgent};
	char fwd[17];
	unsigned i;
#if SUPPORT_COUNTRY
	char* country = id2code(countries, r->country < 0 ? (r->country = ip2id(countries, r->ip)) : r->country);
#else
	char* country = "-";
#endif

	if (r->query) r->query[-1] = '?';
	if (r->params) r->params[-1] = ';';

	for (i = 0; i < sizeof(strings)/sizeof(strings[0]); i++) {
		char* c;
		if ((c = strings[i])) while((c = strchr(c, '\t'))) *c++ = ' ';
	}

	if (r->forwardedFor) {
		*fwd = '-';
		strcpy(fwd + 1, ip2string(r->forwardedFor));
	} else {
		*fwd = 0;
	}
	message(ACCESS, "\t%s%s\t%s\t%s\t%d\t%s\t%s%s\t%s\t%s\t%s",
			ip2string(r->ip), fwd, country, r->cookie?:"-",
			r->status, methods[r->method], r->host?:"-", r->uri?:"-",
			itoa(r->sent), r->referrer?:"-", r->userAgent?:"-");

	if (r->query) r->query[-1] = 0;
	if (r->params) r->params[-1] = 0;
}



int
setRange(Request* r)
{
	if (r->content) return OK;
	if (!r->acceptRanges || r->length < 0 || r->status != 200) {
		r->acceptRanges = NO;
		r->rangeFrom = 0;
		r->rangeTo = 0;
		return OK;
	}

	if (r->rangeFrom >= 0 || r->rangeTo >= 0) {
		r->status = 206;
		if (r->rangeFrom < 0) { // bytes=-500
			if (r->rangeTo < r->length) {
				r->rangeTotal = r->length;
				r->length = r->rangeTo;
				r->rangeTo = r->rangeTotal - 1;
				r->rangeFrom = r->rangeTotal - r->length;
				return OK;
			}
		} else if (r->rangeTo < 0) { // bytes=500-
			if (r->rangeFrom < r->length) {
				r->rangeTotal = r->length;
				r->length = r->rangeTotal - r->rangeFrom;
				r->rangeTo = r->rangeTotal - 1;
				return OK;
			}
		} else if (r->rangeFrom <= r->rangeTo && r->rangeFrom < r->length) {
			// bytes=500-1000
			r->rangeTotal = r->length;
			if (r->rangeTo >= r->rangeTotal)
				r->rangeTo = r->rangeTotal - 1;
			r->length = r->rangeTo - r->rangeFrom + 1;
			return OK;
		}
		r->status = 416;
		r->rangeFrom = r->rangeTo = -1;
		return FAILED;
	}
	r->status = 200;
	r->rangeFrom = 0;
	r->rangeTo = r->length - 1;
    return OK;
}

void
http(Request* r)
{
	char* line;
	char buffer[BLOCK_SIZE];
#if MODEL != MODEL_RTSIG
	int watchRead = 0;
	int watchWrite = 0;
#endif

	currentRequest = r;

	while(1) {
		switch (r->state) {
		case UNDEF:
			debug("%d> http: %s", r->fd, states[r->state]);
			message(ERROR, "strange: http in UNDEF state (%d)", r->fd);
			goto end;

		case READ_REQUEST:
#if MODEL != MODEL_RTSIG
			watchRead = 1;
#endif
		case KEEP_ALIVE:
			debug("%d> http: %s", r->fd, states[r->state]);

			if (!r->buffer) r->buffer = newBuffer(INITIAL_BUFFER_SIZE);

			if (!(line = getLine(r))) {
				if (errno == EWOULDBLOCK) goto end;
				r->state = CLOSE;
				break;
			}
			if (!*line) break; // skip empty lines before request

			r->start = now;
			r->wasRequest = YES;
			removeFromWaiting(r);
			insertToRequests(r);

			if ((r->status = parseRequestLine(line, r))) {
				r->state = NOT_OK;
				break;
			}
			r->state = READ_HEADER;
			//break;

		case READ_HEADER:
			debug("%d> http: %s", r->fd, states[r->state]);
			while (1) {
				if (!(line = getLine(r))) {
					if (errno == EWOULDBLOCK) goto end;
					r->state = CLOSE;
					break;
				}
				if (!*line) {
					int count = r->buffer->count - r->buffer->start;
					r->state = HANDLE_REQUEST;

					if (r->method == POST) {
						if (r->inContentLength < 0) {
							message(MESSAGE, "warning: post without content-length [%s %s%s]", ip2string(r->ip), r->host?:"-", r->uri);
							r->status = 411;
							r->state = NOT_OK;
						} else if (r->inContentLength > config->maxPostSize) {
							message(MESSAGE, "warning: post size too big (%d > %d) [%s %s%s]", r->inContentLength, config->maxPostSize, ip2string(r->ip), r->host?:"-", r->uri);
							r->status = 413;
							r->state = NOT_OK;
						} else if (r->inContentLength) {
							debug("creating postBuffer(%d in %d)", r->inContentLength, count);
							r->postBuffer = newBuffer(r->inContentLength);
							if (count > r->inContentLength) count = r->inContentLength;
							if (count) {
								int pipelined = r->buffer->count - r->buffer->start - count;
								memcpy(r->postBuffer->buffer, r->buffer->buffer + r->buffer->start, r->postBuffer->count = count);
								if (pipelined)
									memmove(r->buffer->buffer + r->buffer->start,
											r->buffer->buffer + r->buffer->start + count,
											pipelined);
								debug("shrinking buffer from %d to %d (pipelined: %d)", r->buffer->count, r->buffer->count - count, pipelined);
								r->buffer->count -= count;
							}

							if (count < r->inContentLength)
								r->state = READ_POST;
						}
					}

					break;
				}

				if ((r->status = parseHeaderLine(line, r))) {
					r->state = NOT_OK;
					break;
				}
			}
			break;

		case READ_POST: {
			ssize_t cnt;
			Buffer* buffer = r->postBuffer;
			debug("%d> http: %s", r->fd, states[r->state]);

			errno = 0;
			cnt = read(r->fd, buffer->buffer + buffer->count, buffer->size - buffer->count);
			debug("%d> reading %d bytes (%s)", r->fd, cnt, errno?strerror(errno):"OK");
			if (cnt <= 0) {
				if (errno == EWOULDBLOCK) goto end;
				if (cnt < 0 && errno != ECONNRESET) message(SYS|ERROR, "cannot read from %s", ip2string(r->ip));
				r->state = CLOSE;
				break;
			}
			buffer->count += cnt;

			if (buffer->count == buffer->size) r->state = HANDLE_REQUEST;

			break;
		}

		case HANDLE_REQUEST: {
			debug("%d> http: %s", r->fd, states[r->state]);

#if MODEL != MODEL_RTSIG
			watchRead = -1;
#endif
			r->pipelined = r->buffer->count - r->buffer->start;

			totalRequests++;

			r->config = config;
			config->active++;

			if ((r->status = handleRequest(r))) {
				r->state = NOT_OK;
				break;
			}
			if (r->state == HANDLING) goto end;

			r->state = PREPARE_HEADER;
			break;
		}

		case HANDLING:
			debug("%d> http: %s", r->fd, states[r->state]);
#if MODEL == MODEL_EPOLL
			if ((lastEventData & EPOLLERR)) {
				debug("%d> error event while handling %s: %s",
					  r->fd, r->uri, showEpollFlags(lastEventData));
				r->state = CLOSE;
				break;
			} else {
				debug("%d> strange: event while handling %s: %s",
					  r->fd, r->uri, showEpollFlags(lastEventData));
			}
#else
			debug("%d> strange: event while handling %s", r->fd, r->uri);
#endif
			goto end;


		case PREPARE_HEADER: {
			debug("%d> http: %s", r->fd, states[r->state]);

			if (r->status == 0) r->status = 200;
			if (r->length < 0 && !(r->method == HEAD || r->status == 304))
				r->keepalive = NO;

			if (r->alias && r->alias->expires) {
				time_t expires =
					r->alias->expires == EXPIRES_PAST ? EXPIRES_PAST :
					r->alias->expires == EXPIRES_FUTURE ? EXPIRES_FUTURE :
					now + r->alias->expires;

				if (!r->expires || expires < r->expires) r->expires = expires;
			}

			if (!setRange(r)) {
				r->state = NOT_OK;
				break;
			}

			r->state = WRITE_HEADER;
#if MODEL != MODEL_RTSIG
			watchWrite = 1;
#endif

			if (config->sendBuffer)
				if (setsockopt(r->fd, SOL_SOCKET, SO_SNDBUF, &config->sendBuffer, sizeof(config->sendBuffer)))
					message(SYS|ERROR, "cannot setsockopt(SO_SNDBUF=%d)", config->sendBuffer);
			//break;
		}

		case WRITE_HEADER: {
			char header[MEDIUM_TEXT];
			ssize_t count = 0;
			ssize_t l;
			ssize_t length;
			ssize_t body = 0;
			struct iovec buffers[2];

			debug("%d> http: %s", r->fd, states[r->state]);

			l = fillHeaders(r, header, MEDIUM_TEXT);

			length = (r->method == HEAD || r->status == 304 || r->status == 204) ? 0 : r->length;

			l -= r->offset;
			buffers[0].iov_base = header + r->offset;
			buffers[0].iov_len = l;
			buffers[1].iov_len = 0;

			if (r->content) {
#if SUPPORT_FILTER
				if (r->contentType && r->alias && r->alias->filter &&
					r->alias->filter->type &&
					!strncmp(r->contentType, r->alias->filter->type, r->alias->filter->typeLength)) {
					r->filter.state = 1;
					r->filter.pos = 0;
					r->filter.quote = 0;
					r->alias->filter->handler(r, NO);
				} else {
					r->filtered = r->received;
				}
				buffers[1].iov_len = body = r->filtered;
#else
				buffers[1].iov_len = body = r->received;
#endif
				buffers[1].iov_base = r->content;
#if SUPPORT_CACHE
				if (r->cached) {
					debug("copy to cached %d bytes", body);
					memcpy(r->cached->body, r->content, body);
				}
#endif


			} else if (length > 0) {
				if (r->body) {
					buffers[1].iov_base = r->body + r->rangeFrom;
					buffers[1].iov_len = body = length;
				} else if (r->file > 0) {
#if TRANSFER_MODE == TRANSFER_PLAIN
					if (r->rangeFrom) lseek(r->file, r->rangeFrom, SEEK_SET);
					if ((body = read(r->file, buffer, BLOCK_SIZE)) < 0) {
						message(SYS|ERROR, "cannot read %s%s", r->server ? r->server->name : "?", r->uri);
						r->state = CLOSE;
						break;
					}
					buffers[1].iov_base = buffer;
					buffers[1].iov_len = body;

#elif TRANSFER_MODE == TRANSFER_SENDFILE
#if defined(HAS_LINUX_SENDFILE)
					int onoff = 1;
					if (setsockopt(r->fd, SOL_TCP, TCP_CORK, &onoff, sizeof(onoff)))
						message(SYS|ERROR, "cannot setsockopt(TCP_CORK=1)");
#elif defined(HAS_BSD_SENDFILE)
					off_t sbytes = -1;
					struct sf_hdtr header;
					header.headers = buffers;
					header.hdr_cnt = 1;
					header.trailers = NULL;
					header.trl_cnt = 0;

					sendfile(r->file, r->fd, r->rangeFrom, length, &header, &sbytes, 0);
					if (sbytes < buffers[0].iov_len) {
						debug("%d> written %d bytes (%s)", r->fd, sbytes, errno?strerror(errno):"OK");
						if (errno == EWOULDBLOCK) {
							debug("%d> blocked header write (%d bytes)", r->fd, buffers[0].iov_len);
							goto end;
						}
						if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot sendfile to %s", ip2string(r->ip));
						r->state = CLOSE;
						break;
					}
					count = sbytes;
					body = length;

#endif

#elif TRANSFER_MODE == TRANSFER_MMAP
					r->mmap = mmap(NULL, r->fileSize, PROT_READ, MAP_PRIVATE, r->file, 0);
					freeFD(r->file);
					r->file = -1;
					if (r->mmap == MAP_FAILED) {
						message(SYS|ERROR, "cannot mmap %s%s", r->server ? r->server->name : "?", r->uri);
						r->state = CLOSE;
						break;
					}
					buffers[1].iov_base = r->mmap + r->rangeFrom;
					buffers[1].iov_len = body = length;
#else
#error unknown TRANSFER_MODE
#endif
				}
			} else {
				r->body = NULL;
			}

			l += body;

			errno = 0;
			if (buffers[1].iov_len) {
				if ((count = writev(r->fd, buffers, 2)) <= 0) {
					debug("%d> written %d bytes (%s)", r->fd, count, errno?strerror(errno):"OK");
					if (errno == EWOULDBLOCK) goto end;
					if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
					r->state = CLOSE;
					break;
				}
			} else if (!count) {
				if ((count = write(r->fd, buffers[0].iov_base, buffers[0].iov_len)) <= 0) {
					debug("%d> written %d bytes (%s)", r->fd, count, errno?strerror(errno):"OK");
					if (errno == EWOULDBLOCK) goto end;
					if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
					r->state = CLOSE;
					break;
				}
			}
			debug("%d> written %d bytes (%s)", r->fd, count, errno?strerror(errno):"OK");

			if (r->server) r->server->sent += count;

			if (count < l) {
				if (count < (ssize_t)buffers[0].iov_len) {
					debug("%d> incomplete header write (%d/%d bytes)", r->fd, count, buffers[0].iov_len);
					r->offset += count;
					goto end;
				}
				body -= l - count;
			}

			r->sent += body;
			r->length = length;

			if (body == r->length) {
				r->state = DONE;
				break;
			}
			r->offset = 0;

			if (r->content) {
				r->state = WRITE_CONTENT;
				if (r->length < 0 || r->contentSize < r->length) {
					if (r->received > r->sent) memmove(r->content, r->content + r->sent, r->received - r->sent);
					r->offset = r->sent;
				}
			} else {
				if (r->body && !r->alloc &&
#if SUPPORT_CACHE
					!r->cached &&
#endif
					!(r->body >= r->buffer->buffer &&
					  r->body < r->buffer->buffer + r->buffer->count))
					r->body = r->alloc = memcpy(MALLOC(r->length + r->rangeFrom), r->body, r->length + r->rangeFrom);
				r->state = WRITE_BODY;
			}

			if (r->content) break;

#if defined(HAS_LINUX_SENDFILE) && TRANSFER_MODE == TRANSFER_SENDFILE
			if (r->file > 0) break;
#endif

			goto end;
		}

		case WRITE_BODY:
			debug("%d> http: %s", r->fd, states[r->state]);
#if SUPPORT_THROTTLE
			if (r->sent && r->server && r->server->sent >= r->server->sentLimit) {
				debug("%d> throttle: %d sent >= %d limit", r->fd, r->server->sent, r->server->sentLimit);
				r->state = THROTTLED;
				insertToThrottled(r->server, r);
#if MODEL != MODEL_RTSIG
				watchWrite = -1;
#endif
				goto end;
			}
#endif /* SUPPORT_THROTTLE */

			if (r->body) {
				ssize_t written;
				errno = 0;
				if ((written = write(r->fd, r->body + r->sent + r->rangeFrom, r->length - r->sent)) <= 0) {
					debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");
					if (errno == EWOULDBLOCK) goto end;
					if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
					r->state = CLOSE;
					break;
				}
				debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");
				r->sent += written;
				if (r->server) r->server->sent += written;
				if (r->sent == r->length) {
					r->state = DONE;
					break;
				}
				goto end;
			} else if (r->file > 0) {
#if TRANSFER_MODE == TRANSFER_PLAIN
				lseek(r->file, r->sent + r->rangeFrom, SEEK_SET);
				while (1) {
					ssize_t written;
					ssize_t readed = r->length - r->sent;
					errno = 0;
					if (readed > BLOCK_SIZE) readed = BLOCK_SIZE;
					if ((readed = read(r->file, buffer, readed)) < 0) {
						message(SYS|ERROR, "cannot read localfile");
						r->state = CLOSE;
						break;
					}
					if (readed == 0) {
						message(SYS|ERROR, "unexpected end of file");
						r->state = CLOSE;
						break;
					}
					if ((written = write(r->fd, buffer, readed)) <= 0) {
						debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");
						if (errno == EWOULDBLOCK) goto end;
						if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
						r->state = CLOSE;
						break;
					}
					debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");

					r->sent += written;
					if (r->server) r->server->sent += written;
					if (written < readed) goto end;
					if (r->sent == r->length) {
						r->state = DONE;
						break;
					}
				}
#elif TRANSFER_MODE == TRANSFER_SENDFILE
				errno = 0;
				{
					off_t offset = r->sent + r->rangeFrom;
					ssize_t sent;
#if defined(HAS_BSD_SENDFILE)
					off_t sbytes = 0;
#endif

#if defined(HAS_LINUX_SENDFILE)
					sent = sendfile(r->fd, r->file, &offset, r->length - r->sent);

					if (r->sent == 0 && !r->pipelined) {
						int onoff = 0;
						if (setsockopt(r->fd, SOL_TCP, TCP_CORK, &onoff, sizeof(onoff)))
							message(SYS|ERROR, "cannot setsockopt(TCP_CORK=0)");
					}
#elif defined(HAS_BSD_SENDFILE)
					sendfile(r->file, r->fd, offset, r->length - r->sent, NULL, &sbytes, 0);
					if (sbytes > 0) sent = sbytes;
					else sent = -1;
#endif

					debug("%d> sendfile %d bytes (%s)", r->fd, sent, errno?strerror(errno):"OK");
					if (sent > 0) {
						r->sent += sent;
						if (r->server) r->server->sent += sent;
						if (r->sent == r->length) {
							r->state = DONE;
							break;
						}
					}
					if (sent <= 0 || errno) {
						if (errno == EWOULDBLOCK) goto end;
						if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot sendfile to %s", ip2string(r->ip));
						r->state = CLOSE;
						break;
					}
					goto end;
				}
#elif TRANSFER_MODE == TRANSFER_MMAP
				message(MESSAGE, "strange, r->file in mmap mode");
#else
#error unknown TRANSFER_MODE
#endif
#if TRANSFER_MODE == TRANSFER_MMAP
			} else if (r->mmap) {
				ssize_t written;
				errno = 0;
				if ((written = write(r->fd, r->mmap + r->rangeFrom + r->sent, r->length - r->sent)) <= 0) {
					debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");
					if (errno == EWOULDBLOCK) goto end;
					if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
					r->state = CLOSE;
					break;
				}
				debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");
				r->sent += written;
				if (r->server) r->server->sent += written;
				if (r->sent == r->length) {
					r->state = DONE;
					break;
				}
				goto end;
#endif // TRANSFER_MMAP
			} else {
				r->state = DONE;
			}
			break;

		case WRITE_CONTENT:
			debug("%d> http: %s", r->fd, states[r->state]);
#if SUPPORT_PROXY


#if SUPPORT_THROTTLE
			if (r->sent && r->server && r->server->sent >= r->server->sentLimit) {
				debug("%d> throttle: %d sent >= %d limit", r->fd, r->server->sent, r->server->sentLimit);
				r->state = THROTTLED;
				insertToThrottled(r->server, r);
#if MODEL != MODEL_RTSIG
				watchWrite = -1;
#endif
				goto end;
			}
#endif /* SUPPORT_THROTTLE */

#if SUPPORT_FILTER
			if (r->filter.state) r->alias->filter->handler(r, NO);
			else r->filtered = r->received;
#endif
			while (1) {
#if SUPPORT_FILTER
				ssize_t count = r->filtered - r->sent;
#else
				ssize_t count = r->received - r->sent;
#endif

				if (count == 0) {
					if (r->proxy && r->proxy->state == WAIT_MEMORY) proxy(r->proxy);
#if MODEL != MODEL_RTSIG
					watchWrite = -1;
#endif
					r->state = WAIT_DATA;
					break;
				}
				debug("%d> WRITE_CONTENT: size: %d, off: %d, r: %d, s: %d", r->fd, r->contentSize, r->offset, r->received, r->sent);
				if ((count = write(r->fd, r->content + r->sent - r->offset, count)) <= 0) {
					if (errno == EWOULDBLOCK) {
						debug("%d> blocked write", r->fd);
						if (r->proxy && r->proxy->state == WAIT_MEMORY) proxy(r->proxy);
						goto end;
					}
					if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
					r->state = CLOSE;
					break;
				}
				debug("%d> writing %d bytes", r->fd, count);
#if SUPPORT_CACHE
				if (r->cached) {
                    debug("%d> copy to cached %d bytes from %d", r->fd, count, r->sent);
					memcpy(r->cached->body + r->sent, r->content + r->sent - r->offset, count);
				}
#endif
				r->sent += count;
				if (r->server) r->server->sent += count;
#if SUPPORT_FILTER
				if (r->filtered == r->sent && (!r->proxy || r->proxy->state == FINISHED)) {
					if (r->filtered == r->received) {
						r->state = DONE;
						break;
					} else {
						r->alias->filter->handler(r, YES);
						continue;
					}
				}
#else
				if (r->received == r->sent && (!r->proxy || r->proxy->state == FINISHED)) {
					r->state = DONE;
					break;
				}
#endif
				if (r->length < 0 || r->offset + r->contentSize < r->length) {
					if (r->received > r->sent) memmove(r->content, r->content + r->sent - r->offset, r->received - r->sent);
					r->offset = r->sent;
				}
			}
#endif /* SUPPORT_PROXY */
			break;


		case WAIT_DATA:
			debug("%d> http: %s", r->fd, states[r->state]);
#if SUPPORT_PROXY
#if SUPPORT_FILTER
			if (r->filter.state) r->alias->filter->handler(r, NO);
			else r->filtered = r->received;
			if (r->filtered == r->sent)
#else
				if (r->received == r->sent)
#endif
				{
					if (!r->proxy || r->proxy->state == FINISHED) {
#if SUPPORT_FILTER
						if (r->filtered < r->received) {
							r->alias->filter->handler(r, YES);
							r->state = WRITE_CONTENT;
#if MODEL != MODEL_RTSIG
							watchWrite = 1;
#endif
						} else	{
							r->state = DONE;
						}
#else
						r->state = DONE;
#endif
					} else {
						goto end;
					}
				} else {
					r->state = WRITE_CONTENT;
#if MODEL != MODEL_RTSIG
					watchWrite = 1;
#endif
				}
#endif /* SUPPORT_PROXY */
			break;

#if SUPPORT_THROTTLE
		case THROTTLED:
			debug("%d> http: %s", r->fd, states[r->state]);

			removeFromThrottled(r->server, r);

#if MODEL != MODEL_RTSIG
			watchWrite = 1;
#endif
			r->state = r->content ? WRITE_CONTENT : WRITE_BODY;
            break;

#endif /* SUPPORT_THROTTLE */


		case DONE: {
#if defined(HAS_LINUX_SENDFILE) && TRANSFER_MODE == TRANSFER_SENDFILE
			int wasFile = r->file > 0 ? YES : NO;
#endif
			debug("%d> http: %s", r->fd, states[r->state]);
			if (r->mmap) {
				munmap(r->mmap, r->fileSize);
				r->mmap = NULL;
			}
			if (r->file > 0 && freeFD(r->file) < 0)
				message(SYS|ERROR, "cannot close localfile");
			r->file = -1;
			if (r->postBuffer) r->postBuffer = freeBuffer(r->postBuffer);
			if (r->alloc) r->alloc = FREE(r->alloc);
#if SUPPORT_CACHE
			if (r->cached) {
				if (r->content) activateEntry(r->alias->cache, r->cached);
				r->cached = unlinkEntry(r->alias->cache, r->cached);
			}
#endif
#if SUPPORT_PROXY
			if (r->proxy) closeProxy(r->proxy);
			if (r->contentSize) {
				r->content = FREE(r->content);
				r->contentSize = 0;
			}
#endif

			if (r->keepalive) {
				unsigned ip = r->ip;
				int fd = r->fd;
				Buffer* buffer = r->buffer;
				int pipelined = buffer->count - buffer->start;
				char* s;
				Request *prev, *next;
#if SUPPORT_COUNTRY
                short country = r->country;
                short region = r->region;
#endif
#if SUPPORT_ACCESS
				char trusted = r->trusted;
#endif

#if MODEL != MODEL_RTSIG
				char wRead, wWrite;
#endif

				int l = r->wasRequest ?
					r->alias ? r->alias->log :
					r->server ? r->server->log :
					config->log : NO;
				if (l == OK || (l == ERRORS && r->status >= 400)) logRequest(r);

				for (s = buffer->buffer + buffer->start; pipelined; s++, pipelined--)
					if (*s != '\r' && *s != '\n') break;

				if (pipelined) {
					memmove(buffer->buffer, s, pipelined);
					buffer->count = pipelined;
					buffer->start = 0;
				} else {
					buffer = freeBuffer(buffer);
				}

				if (r->config && (--(r->config->active)) == 0 && r->config != config)
					freeConfig(r->config);

#if MODEL != MODEL_RTSIG
				wRead = r->watchRead;
				wWrite = r->watchWrite;
#endif
				if (r->wasRequest) {
					removeFromRequests(r);
					insertToWaiting(r);
				} else {
					message(MESSAGE, "wow! no request, but in DONE");
				}

				prev = r->prev;
				next = r->next;


				memset(r, 0, sizeof(Request));

				r->type = REQUEST;
				r->state = KEEP_ALIVE;
				r->fd = fd;
				r->major = 1;
				r->rangeFrom = r->rangeTo = r->file =
					r->length = r->inContentLength = -1;
				r->ip = ip;
#if SUPPORT_COUNTRY
				r->country = country;
				r->region = region;
#endif
#if SUPPORT_ACCESS
				r->trusted = trusted;
#endif
				r->last = now + config->keepAliveTimeout;
				r->start = now;
				r->buffer = buffer;
#if MODEL != MODEL_RTSIG
				r->watchRead = wRead;
				r->watchWrite = wWrite;
#endif

				r->prev = prev;
				r->next = next;

#if MODEL != MODEL_RTSIG
				watchRead = 1;
				watchWrite = -1;
#endif

				if (!pipelined) {
#if defined(HAS_LINUX_SENDFILE) && TRANSFER_MODE == TRANSFER_SENDFILE
					int onoff = 0;
					if (wasFile && setsockopt(r->fd, SOL_TCP, TCP_CORK, &onoff, sizeof(onoff)))
						message(SYS|ERROR, "cannot setsockopt(TCP_CORK=0)");
#endif
					goto end;
				}

				break;
			}

			r->state = CLOSE;
			//break;
		}

		case CLOSE: {
			int l = r->wasRequest ?
				r->alias ? r->alias->log :
				r->server ? r->server->log :
				config->log :
				NO;

			debug("%d> http: %s", r->fd, states[r->state]);

			if (l == YES || (l == ERRORS && r->status >= 400)) logRequest(r);

#if SUPPORT_ANTIFLOOD
#if SUPPORT_ACCESS
			if (!r->trusted)
#endif
				decIP(r->ip);
#endif /* SUPPORT_ANTIFLOOD */

#if SUPPORT_THROTTLE
			if (r->throttled) removeFromThrottled(r->server, r);
#endif


			if (r->mmap) munmap(r->mmap, r->fileSize);
			if (r->file > 0 && freeFD(r->file) < 0)
				message(SYS|ERROR, "cannot close localfile");
			if (r->fd > 0) {
#if MODEL == MODEL_KEVENT
                if (r->watchRead || r->watchWrite) unwatchEvent(r->fd);
#endif
				fds[r->fd].request = NULL;
				// some weirdness: if any bytes in socket buffer
				// connection reset appears,
				// quick-fix: read them all
				if (r->method == POST) while(read(r->fd, buffer, BLOCK_SIZE) == BLOCK_SIZE);
				if (close(r->fd) < 0)
					message(SYS|ERROR, "cannot close %d", r->fd);
			}
			if (r->buffer) freeBuffer(r->buffer);
			if (r->postBuffer) freeBuffer(r->postBuffer);
			if (r->alloc) FREE(r->alloc);
#if SUPPORT_CACHE
			if (r->cached) r->cached = unlinkEntry(r->alias->cache, r->cached);
#endif
#if SUPPORT_PROXY
			if (r->proxy) closeProxy(r->proxy);
			if (r->contentSize) FREE(r->content);
#endif
			if (r->config && (--(r->config->active)) == 0 && r->config != config)
				freeConfig(r->config);

			if (r->wasRequest) removeFromRequests(r);
			else removeFromWaiting(r);

			memset(r, 0, sizeof(Request));

			r->next = freeRequests;
			freeRequests = r;
			nRequests--;
            //showList();
			r = NULL;
			goto end;
		}

		case NOT_OK:
			debug("%d> http: %s", r->fd, states[r->state]);
			if (r->status != 304 && r->status != 204 && !r->body) {
                r->contentType = r->contentEnc = NULL;

				if (r->method != HEAD)
					r->length = makeErrorReply(r, r->body = buffer, BLOCK_SIZE);
				if (!r->contentType)
					r->contentType = getPair(config->extensions, "html") ? : "text/html";
			}
			if (r->status >= 400) r->keepalive = NO;
			r->state = PREPARE_HEADER;
			break;
		}
	}

end:

	if (r) {
#if MODEL == MODEL_KEVENT
		if (watchRead > 0) {
			if (!r->watchRead) {
				watchEvent(r->fd, EVFILT_READ, EV_ADD, 0, 0);
				r->watchRead = YES;
			}
		} else if (watchRead < 0) {
			if (r->watchRead) {
				watchEvent(r->fd, EVFILT_READ, EV_DELETE, 0, 0);
				r->watchRead = NO;
			}
		}

		if (watchWrite > 0) {
			if (!r->watchWrite) {
#if LOWAT_MODE == LOWAT_KEVENT
				if (config->lowWatermark) {
					debug("%d> setting low watermark to %d", r->fd, config->lowWatermark);
					watchEvent(r->fd, EVFILT_WRITE, EV_ADD, NOTE_LOWAT, config->lowWatermark);
				} else {
					watchEvent(r->fd, EVFILT_WRITE, EV_ADD, 0, 0);
				}
#else
				watchEvent(r->fd, EVFILT_WRITE, EV_ADD, 0, 0);
#endif
				r->watchWrite = YES;
			}
		} else if (watchWrite < 0) {
			if (r->watchWrite) {
				watchEvent(r->fd, EVFILT_WRITE, EV_DELETE, 0, 0);
				r->watchWrite = NO;
			}
		}
#elif MODEL == MODEL_EPOLL
		{
            unsigned oldFlag = 0, newFlag = 0;
            if (r->watchRead) oldFlag = EPOLLIN;
            if (r->watchWrite) oldFlag |= EPOLLOUT;

			if (watchRead) r->watchRead = watchRead > 0 ? YES : NO;
			if (watchWrite) r->watchWrite = watchWrite > 0 ? YES : NO;

			if (r->watchRead) newFlag = EPOLLIN;
			if (r->watchWrite) newFlag |= EPOLLOUT;

            if (oldFlag != newFlag)
				watchEvent(r->fd, newFlag, EPOLL_CTL_MOD);
		}
#elif MODEL == MODEL_POLL
		if (watchRead) r->watchRead = watchRead > 0 ? YES : NO;
		if (watchWrite) r->watchWrite = watchWrite > 0 ? YES : NO;
#endif //MODEL


#if LOWAT_MODE == LOWAT_SOCKOPT
		debug("%d> setting low watermark to %d", r->fd, config->lowWatermark);
		if (setsockopt(r->fd, SOL_SOCKET, SO_SNDLOWAT, &config->lowWatermark, sizeof(config->lowWatermark)))
			message(SYS|ERROR, "cannot setsockopt(SO_SNDLOWAT=%d)", config->lowWatermark);
#endif

		if (r->wasRequest) {
			r->last = now + config->timeout;
			if (requests != r) {
				removeFromRequests(r);
				insertToRequests(r);
			}
		}
	}

	currentRequest = NULL;
}


#define split_sec(sec, h, m, s) {\
	h = sec/60/60;\
	m = (sec%(60*60))/60;\
	s = (sec%60);\
	}

int
handleStatusRequest(Request* r)
{
	char* body;
	int real = 0, system = 0, user = 0;
	int h, m, s, l;
	time_t finish;
	struct rusage ru;
	Listener* listener;
	Server* server;
	int size = r->query ? HUGE_TEXT : LARGE_TEXT;
	char show;

	unsigned sumCount = 0, sumRequestsLast = 0, sumSentLast = 0;
	unsigned sumRequestsToday = 0, sumDroppedS = 0, sumDroppedC = 0;
#if SUPPORT_PROXY
	unsigned sumActive = 0;
#endif
	long long sumSentToday = 0;


	if (!(body = MALLOC(size))) return 500;
	l = SNPRINTF(body, size,
				 "<html><head><title>%s</title></head>\n<body><h3>%s</h3><pre>\n"
				 "Server version: %s, built: %s\n"
				 "Current time: %s\n"
				 "Started at    %s\n"
				 "Configured at %s\n",
				 "Server status", "Server status",
				 VERSION, built,
				 time2str(now, LOCAL_TIME),
				 time2str(start, LOCAL_TIME),
				 time2str(configured, LOCAL_TIME)
				);

	finish = time(NULL);
	real = finish - start;
	l += SNPRINTF(body + l, size - l, "Elapsed time:\n");
	split_sec(real, h, m, s);
	l += SNPRINTF(body + l, size - l, "  real:  %4d:%02d:%02d\n", h, m, s);

	if (getrusage(RUSAGE_SELF, &ru)) {
		message(SYS|ERROR, "cannot getrusage");
	} else {
		user = ru.ru_utime.tv_sec + (ru.ru_utime.tv_usec < 500000 ? 0 : 1);
		system = ru.ru_stime.tv_sec + (ru.ru_stime.tv_usec < 500000 ? 0 : 1);
		split_sec(user, h, m, s);
		l += SNPRINTF(body + l, size - l, "  user:   %3d:%02d:%02d (%3.1f%%)\n", h, m, s, user*100.0/real);

		split_sec(system, h, m, s);
		l += SNPRINTF(body + l, size - l, "  system: %3d:%02d:%02d (%3.1f%%)\n", h, m, s, system*100.0/real);
	}

	l += SNPRINTF(body + l, size - l, "\nLast 5 seconds: %s acc/sec, %s req/sec, CPU usage: %3.1f%% user, %3.1f%% system\n", itos((lastAccepts*2+5)/10), itos((lastRequests*2+5)/10), lastUser/50.0, lastSys/50.0);

	m = strlen("99 listeners");
	for(listener = config->listeners; listener; listener = listener->next) {
		s = strlen(listener->addr.s_addr ? inet_ntoa(listener->addr): "*") + 1 + strlen(itoa(listener->port));
		if (s > m) m = s;
	}
	l += SNPRINTF(body + l, size - l,
				  "\n%-*s %5s %10s %10s %10s\n",
				  m + 2, "Listener", "acc/s", "acc/today", "srv-drops", "cli-drops");

	for(listener = config->listeners; listener; listener = listener->next) {
		char name[SHORT_STR];
		SNPRINTF(name, SHORT_STR, "%s:%d",
				 listener->addr.s_addr ? inet_ntoa(listener->addr) : "*",
				 listener->port);

		if (listener->error)
			l += SNPRINTF(body + l, size - l, "  %s%-*s %-38s%s\n",
						  listener->next || !sumCount ? "" : "<u>",
						  m, name,
						  listener->error,
						  listener->next || !sumCount ? "" : "</u>"
						 );
		else
			l += SNPRINTF(body + l, size - l, "  %s%-*s %5s %10s %10s %10s%s\n",
						  listener->next || !sumCount ? "" : "<u>",
						  m, name,
						  itos((listener->acceptsLast*2+5)/10),
						  itos(listener->acceptsToday),
						  itos(listener->droppedS),
						  itos(listener->droppedC),
						  listener->next || !sumCount ? "" : "</u>"
						 );
		sumCount++;
		sumRequestsLast += listener->acceptsLast;
		sumRequestsToday += listener->acceptsToday;
		sumDroppedS += listener->droppedS;
		sumDroppedC += listener->droppedC;
	}
	if (sumCount > 1)
		l += SNPRINTF(body + l, size - l, "  %d %-*s %5s %10s %10s %10s\n",
					  sumCount,
					  m - (sumCount < 10 ? 2 : sumCount < 100 ? 3 : 4), "listeners",
					  itos((sumRequestsLast*2+5)/10),
					  itos(sumRequestsToday),
					  itos(sumDroppedS),
					  itos(sumDroppedC)
					 );

	sumCount = 0; sumRequestsLast = 0; sumRequestsToday = 0;

	m = strlen("99 servers");
	for(server = config->servers; server; server = server->next) {
		s = strlen(server->name);
		if (s > m) m = s;
	}
	l += SNPRINTF(body + l, size - l,
				  "\n%-*s %5s %7s %11s %11s\n",
				  m + 2, "Server", "req/s", "bytes/s", "req/today", "bytes/today");

	for(server = config->servers; server; server = server->next) {
		l += SNPRINTF(body + l, size - l, "  %s%-*s %5s %7s %11s %11s%s\n",
					  server->next || !sumCount ? "" : "<u>",
					  m, server->name,
					  itos((server->requestsLast*2+5)/10),
					  humanSize((server->sentLast*2+5)/10),
					  itos(server->requestsToday),
					  humanSize(server->sentToday),
					  server->next || !sumCount ? "" : "</u>"
					 );
		sumCount++;
		sumRequestsLast += server->requestsLast;
		sumSentLast += server->sentLast;
		sumRequestsToday += server->requestsToday;
		sumSentToday += server->sentToday;
	}

	if (sumCount > 1)
		l += SNPRINTF(body + l, size - l, "  %d %-*s %5s %7s %11s %11s\n",
					  sumCount,
					  m - (sumCount < 10 ? 2 : sumCount < 100 ? 3 : 4), "servers",
					  itos((sumRequestsLast*2+5)/10),
					  humanSize((sumSentLast*2+5)/10),
					  itos(sumRequestsToday),
					  humanSize(sumSentToday)
					 );

#if SUPPORT_PROXY
	if (config->backends) {
		Backend* backend;
		int m = strlen("99 backends");
		sumCount = 0; sumRequestsLast = 0; sumRequestsToday = 0;

		for(backend = config->backends; backend; backend = backend->next) {
			int n = strlen(backend->name) + 1 + strlen(itoa(backend->port));
			if (n > m) m = n;
		}

		l += SNPRINTF(body + l, size - l, "\n"
					  "%-*s state active max req/s req/today\n", m + 2, "Backend");
		for(backend = config->backends; backend; backend = backend->next) {
			char name[MEDIUM_STR];
			int n = strlen(backend->name) + 1 + strlen(itoa(backend->port));
			SNPRINTF(name, MEDIUM_STR, "<span title='%s'>%s:%d</span>",
                     inet_ntoa(backend->addr), backend->name, backend->port);
			l += SNPRINTF(body + l, size - l, "  %s%s%*s %5s %6d %3d %5s %9s%s\n",
						  backend->next || !sumCount ? "" : "<u>",
						  name, m - n, "", backend->state ? "UP" : "DOWN",
						  backend->nRequests, backend->maxRequests,
						  itos((backend->requestsLast*2+5)/10),
						  itos(backend->requestsToday),
						  backend->next || !sumCount ? "" : "</u>"
						 );
			sumCount++;
			sumActive += backend->nRequests;
			sumRequestsLast += backend->requestsLast;
			sumRequestsToday += backend->requestsToday;
		}

		if (sumCount > 1)
			l += SNPRINTF(body + l, size - l, "  %d %-*s %5s %6s %3s %5s %9s\n",
						  sumCount,
						  m - (sumCount < 10 ? 2 : sumCount < 100 ? 3 : 4), "backends",
						  "",
						  itos(sumActive),
						  "",
						  itos((sumRequestsLast*2+5)/10),
						  itos(sumRequestsToday)
						 );
	}

#endif
#if SUPPORT_CACHE
	l += getCacheStats(body + l, size - l, r);
#endif

	show = r->query && strstr(r->query, "requests");
	l += SNPRINTF(body + l, size - l, "\n"
				  "<a href=\"%s%s\">Requests</a> (active/max): %s / %s\n",
				  r->uri, show ? "" : "?requests",
				  itos(nRequests), itos(maxRequests));

	if (show && (requests || waiting)) {
		Request* r;
		l += SNPRINTF(body + l, size - l,
					  "%-14s %-15s %4s [%5s/%-5s] %s\n",
					  "State", "IP", "Time", "Sent", "Total", "URL");
		for (r = requests; r && l < size - 100; r = r->next) {
			if (r->buffer && r->buffer->start)
				l += SNPRINTF(body + l, size - l,
							  "%-14s %-15s %4s [%5s/%-5s] %s%s%s%s\n",
							  states[r->state], ip2string(r->ip),
							  humanPeriod(now - r->start),
							  humanSize(r->sent),
							  r->length < 0 ? "?" : humanSize(r->length),
							  r->host?:"-", r->uri?:"",
							  r->query?"?":"", r->query?:"");
			else
				l += SNPRINTF(body + l, size - l, "%-14s %-15s %4s\n",
							  states[r->state], ip2string(r->ip),
							  humanPeriod(now - r->start));
		}

		for (r = waiting; r && l < size - 100; r = r->next) {
			l += SNPRINTF(body + l, size - l, "%-14s %-15s %4s\n",
						  states[r->state], ip2string(r->ip),
						  humanPeriod(now - r->start));
		}
	}

#if SUPPORT_PROXY
	l += getProxyStats(body + l, size - l, r);
#endif

#if SUPPORT_ANTIFLOOD
	l += getIPStats(body + l, size - l, r);
#endif
	l += getFilesStats(body + l, size - l, r);
	l += SNPRINTF(body + l, size - l, "</pre></body></html>\n");


	r->body = r->alloc = body;
	r->length = l < size ? l : size;
	r->expires = r->modified = now;
	r->contentType =
		getPair(r->server->extensions, "html") ? :
		getPair(config->extensions, "html") ? :
		"text/html";

	return 0;
}


